# @file test_az_cli_universal_dependency.py
# Unit test suite for the Azure CLI Artifacts Universal Packages Dependency class.
#
# NOTE: To run most of this test you must specify a valid PAT that can read packages from
# tianocore devops (https://dev.azure.com/tianocore) in your environment as PAT_FOR_UNIVERSAL_ORG_TIANOCORE
#
# Universal Packages do not support anonymous access (even in public feeds) and therefore this unit test can not run
# in a GitHub PR flow because it would need secrets.
# https://docs.microsoft.com/en-us/azure/devops/artifacts/concepts/feeds?view=azure-devops#public-feeds
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import os
import unittest
import logging
import tempfile
from edk2toolext.environment import environment_descriptor_files as EDF
from edk2toolext.environment.extdeptypes.az_cli_universal_dependency import AzureCliUniversalDependency
from edk2toolext.environment import version_aggregator
from edk2toollib.utility_functions import RemoveTree

test_dir = None

single_file_json_template = '''
{
  "scope": "global",
  "type": "az-universal",
  "name": "hello-world-file",
  "source": "https://dev.azure.com/tianocore",
  "project": "edk2-pytool-extensions",
  "version": "%s",
  "feed": "ext_dep_unit_test_feed",
  "pat_var": "PAT_FOR_UNIVERSAL_ORG_TIANOCORE"
}
'''

folders_json_template = '''
{
  "scope": "global",
  "type": "az-universal",
  "name": "hello-world-folders",
  "source": "https://dev.azure.com/tianocore",
  "project": "edk2-pytool-extensions",
  "version": "%s",
  "feed": "ext_dep_unit_test_feed",
  "pat_var": "PAT_FOR_UNIVERSAL_ORG_TIANOCORE"
}
'''

file_filter_json_template = '''
{
  "scope": "global",
  "type": "az-universal",
  "name": "hello-world-folders",
  "source": "https://dev.azure.com/tianocore",
  "project": "edk2-pytool-extensions",
  "version": "0.2.1",
  "feed": "ext_dep_unit_test_feed",
  "file-filter": "folder2/*.txt",
  "pat_var": "PAT_FOR_UNIVERSAL_ORG_TIANOCORE"
}
'''


def prep_workspace():
    global test_dir
    # if test temp dir doesn't exist
    if test_dir is None or not os.path.isdir(test_dir):
        test_dir = tempfile.mkdtemp()
        logging.debug("temp dir is: %s" % test_dir)
    else:
        clean_workspace()
        test_dir = tempfile.mkdtemp()


def clean_workspace():
    global test_dir
    if test_dir is None:
        return

    if os.path.isdir(test_dir):
        RemoveTree(test_dir)
        test_dir = None


class TestAzCliUniversalDependency(unittest.TestCase):
    def setUp(self):
        prep_workspace()

    @classmethod
    def setUpClass(cls):
        logger = logging.getLogger('')
        logger.addHandler(logging.NullHandler())
        unittest.installHandler()

    @classmethod
    def tearDownClass(cls):
        clean_workspace()

    def tearDown(self):
        # we need to reset the version aggregator each time
        version_aggregator.GetVersionAggregator().Reset()

    # good case
    @unittest.skipIf("PAT_FOR_UNIVERSAL_ORG_TIANOCORE" not in os.environ.keys(),
                     "PAT not defined therefore universal packages tests will fail")
    def test_download_good_universal_dependency_single_file(self):
        version = "0.0.1"
        ext_dep_file_path = os.path.join(test_dir, "unit_test_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(single_file_json_template % version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = AzureCliUniversalDependency(ext_dep_descriptor)
        ext_dep.fetch()
        self.assertTrue(ext_dep.verify())
        self.assertEqual(ext_dep.version, version)
        # make sure we clean up after ourselves
        ext_dep.clean()

    # good case
    @unittest.skipIf("PAT_FOR_UNIVERSAL_ORG_TIANOCORE" not in os.environ.keys(),
                     "PAT not defined therefore universal packages tests will fail")
    def test_download_good_universal_dependency_folders_pinned_old_version(self):
        version = "0.2.0"
        ext_dep_file_path = os.path.join(test_dir, "unit_test_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(folders_json_template % version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = AzureCliUniversalDependency(ext_dep_descriptor)
        ext_dep.fetch()
        self.assertTrue(ext_dep.verify())
        self.assertEqual(ext_dep.version, version)
        # make sure we clean up after ourselves
        ext_dep.clean()

    # good case
    @unittest.skipIf("PAT_FOR_UNIVERSAL_ORG_TIANOCORE" not in os.environ.keys(),
                     "PAT not defined therefore universal packages tests will fail")
    def test_download_good_universal_dependency_folders_newer_version(self):
        version = "0.2.1"
        ext_dep_file_path = os.path.join(test_dir, "unit_test_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(folders_json_template % version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = AzureCliUniversalDependency(ext_dep_descriptor)
        ext_dep.fetch()
        self.assertTrue(ext_dep.verify())
        self.assertEqual(ext_dep.version, version)
        # make sure we clean up after ourselves
        ext_dep.clean()

    # good case
    @unittest.skipIf("PAT_FOR_UNIVERSAL_ORG_TIANOCORE" not in os.environ.keys(),
                     "PAT not defined therefore universal packages tests will fail")
    def test_download_good_universal_dependency_folders_file_filter(self):
        version = "0.2.1"
        ext_dep_file_path = os.path.join(test_dir, "unit_test_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(file_filter_json_template)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = AzureCliUniversalDependency(ext_dep_descriptor)
        ext_dep.fetch()
        self.assertTrue(ext_dep.verify())
        self.assertEqual(ext_dep.version, version)

        # make sure we have 1 folder and 2 file (ext_Dep state file plus our 1 file from package)

        files = 0
        folders = 0

        for (dirpath, dirs, file_names) in os.walk(ext_dep.contents_dir):
            files += len(file_names)
            folders += len(dirs)

        self.assertEqual(folders, 1)
        self.assertEqual(files, 2)

        # make sure we clean up after ourselves
        ext_dep.clean()

    # bad case
    @unittest.skipIf("PAT_FOR_UNIVERSAL_ORG_TIANOCORE" not in os.environ.keys(),
                     "PAT not defined therefore universal packages tests will fail")
    def test_download_bad_universal_dependency(self):
        non_existing_version = "0.1.0"
        ext_dep_file_path = os.path.join(test_dir, "unit_test_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(single_file_json_template % non_existing_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = AzureCliUniversalDependency(ext_dep_descriptor)
        with self.assertRaises(Exception):
            ext_dep.fetch()
        self.assertFalse(ext_dep.verify())

    def test_az_tool_environment(self):
        AzureCliUniversalDependency.VerifyToolDependencies()


if __name__ == '__main__':
    unittest.main()



# @file shell_environment.py
# This module contains code that helps to manage the build environment
# including PATH, PYTHONPATH, and ENV variables.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code that helps manage and maintain the build environment.

This management includes PATH, PYTHONPATH and ENV Variables.
"""
import os
import sys
import copy
import logging
from edk2toolext.environment import var_dict

LOGGING_GROUP = "EnvDict"
MY_LOGGER = logging.getLogger(LOGGING_GROUP)


#
# Copy the Singleton pattern from...
#   https://stackoverflow.com/a/6798042
#
class Singleton(type): # noqa
    _instances = {}

    def __call__(cls, *args, **kwargs): # noqa
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class ShellEnvironment(metaclass=Singleton):
    """An active copy of the current OS environment.

    Allows for easy manipulation of the environment including taking
    screenshots (checkpoints) that are stored and can be accessed
    later.
    """
    # Easy definition for the very first checkpoint
    # when the environment is first created.
    INITIAL_CHECKPOINT = 0

    def __init__(self):
        """Inits the local environment with the initial os environment."""
        # Add all of our logging to the EnvDict group.
        self.logger = logging.getLogger(LOGGING_GROUP)

        # Initialize all other things.
        self.active_environ = None
        self.active_path = None
        self.active_pypath = None
        self.active_buildvars = var_dict.VarDict()
        self.checkpoints = []

        # Grab a copy of the environment as it exists.
        self.import_environment()

        # Create the initial checkpoint.
        self.checkpoint()

    #
    # Management methods.
    # These methods manage the singleton, the surrounding environment, and checkpoints.
    #
    def import_environment(self):
        """Loads the local environment with os environment."""
        # Create a complete copy of os.environ
        self.active_environ = dict()
        for key, value in os.environ.items():
            self.active_environ[key] = value

        # Record the PATH elements of the current environment.
        path = self.active_environ.get('PATH', "")

        # Filter removes empty elements.
        # List creates an actual list rather than a generator.
        # Filter removes empty strings.
        self.active_path = list(filter(None, path.split(os.pathsep)))

        # Record the PYTHONPATH elements of the current environment.
        # When reading PYTHONPATH, try reading the live path from sys.
        self.active_pypath = sys.path

        # Remove PATH and PYTHONPATH from environ copy to force use of active_path and active_pypath
        self.active_environ.pop("PATH", None)
        self.active_environ.pop("PYTHONPATH", None)

    def export_environment(self):
        """Exports enviornment to the OS."""
        # Purge all keys that aren't in the export.
        for key, value in os.environ.items():
            if key not in self.active_environ:
                os.environ.pop(key)

        # Export all internal keys.
        for key, value in self.active_environ.items():
            os.environ[key] = value

        # Set the PATH and PYTHONPATH vars.
        os.environ["PATH"] = os.pathsep.join(self.active_path)
        os.environ["PYTHONPATH"] = os.pathsep.join(self.active_pypath)

        sys.path = self.active_pypath

    def log_environment(self):
        """Logs the current environment to the logger."""
        self.logger.debug("FINAL PATH:")
        self.logger.debug(", ".join(self.active_path))

        self.logger.debug("FINAL PYTHONPATH:")
        self.logger.debug(", ".join(self.active_pypath))

        self.logger.debug("FINAL ENVIRON:")
        environ_list = []
        for key, value in self.active_environ.items():
            environ_list.append("({0}:{1})".format(key, value))
        self.logger.debug(", ".join(environ_list))

    def checkpoint(self):
        """Creates a checkpoint in time.

        Checkpoint stores the following:
        1. active_environment
        2. active_path
        3. active_pypath
        4. active_buildvars
        """
        new_index = len(self.checkpoints)
        self.checkpoints.append({
            'environ': copy.copy(self.active_environ),
            'path': self.active_path,
            'pypath': self.active_pypath,
            'buildvars': copy.copy(self.active_buildvars)
        })

        return new_index

    def restore_checkpoint(self, index):
        """Restore a specific checkpoint."""
        if index < len(self.checkpoints):
            check_point = self.checkpoints[index]
            self.active_environ = copy.copy(check_point['environ'])
            self.active_path = check_point['path']
            self.active_pypath = check_point['pypath']
            self.active_buildvars = copy.copy(check_point['buildvars'])

            self.export_environment()

        else:
            raise IndexError("Checkpoint %s does not exist" % index)

    def restore_initial_checkpoint(self):
        """Restore the initial checkpoint made."""
        self.restore_checkpoint(ShellEnvironment.INITIAL_CHECKPOINT)

    #
    # Environment manipulation methods.
    # These methods interact with the current environment.
    #
    def _internal_set_path(self, path_elements):
        self.active_path = list(path_elements)
        os.environ["PATH"] = os.pathsep.join(self.active_path)

    def _internal_set_pypath(self, path_elements):
        self.active_pypath = list(path_elements)
        os.environ["PYTHONPATH"] = os.pathsep.join(self.active_pypath)
        sys.path = self.active_pypath

    def set_path(self, new_path):
        """Set the path.

        Args:
            new_path (str): path to override with
        """
        self.logger.debug("Overriding PATH with new value.")
        if type(new_path) is str:
            new_path = list(new_path.split(os.pathsep))
        self._internal_set_path(new_path)

    def set_pypath(self, new_path):
        """Set the pypath.

        Args:
            new_path (str): path to override with
        """
        self.logger.debug("Overriding PYTHONPATH with new value.")
        if type(new_path) is str:
            new_path = list(new_path.split(os.pathsep))
        self._internal_set_pypath(new_path)

    def append_path(self, path_element):
        """Append to the end of path.

        if path_element already exists within path it will be removed
        from the current location and appended to the end

        Args:
            path_element (str): path element to append
        """
        self.logger.debug("Appending PATH element '%s'." % path_element)
        if path_element in self.active_path:
            # remove so we don't have duplicates but we respect the order
            # requested by the caller
            self.active_path.remove(path_element)
        self._internal_set_path(self.active_path + [path_element])

    def insert_path(self, path_element):
        """Insert at front of the path.

        if path_element already exists within path it will be removed
        from the current location and prepended to the front

        Args:
            path_element (str): path element to insert
        """
        self.logger.debug("Inserting PATH element '%s'." % path_element)
        if path_element in self.active_path:
            # remove so we don't have duplicates but we respect the order
            # requested by the caller
            self.active_path.remove(path_element)
        self._internal_set_path([path_element] + self.active_path)

    def append_pypath(self, path_element):
        """Append to the end of pypath.

        if path_element already exists within pypath it will be removed
        from the current location and appended to the end

        Args:
            path_element (str): path element to append
        """
        self.logger.debug("Appending PYTHONPATH element '%s'." % path_element)
        if path_element in self.active_pypath:
            # remove so we don't have duplicates but we respect the order
            # requested by the caller
            self.active_pypath.remove(path_element)
        self._internal_set_pypath(self.active_pypath + [path_element])

    def insert_pypath(self, path_element):
        """Insert at front of the pypath.

        if path_element already exists within pypath it will be removed
        from the current location and prepended to the front

        Args:
            path_element (str): path element to insert
        """
        self.logger.debug("Inserting PYTHONPATH element '%s'." % path_element)
        if path_element in self.active_pypath:
            # remove so we don't have duplicates but we respect the order
            # requested by the caller
            self.active_pypath.remove(path_element)
        self._internal_set_pypath([path_element] + self.active_pypath)

    def replace_path_element(self, old_path_element, new_path_element):
        """Replaces the PATH element.

        Generates a new PATH by iterating through the old PATH and replacing
        old_path_element with new_path_element where it is found.

        Args:
            old_path_element (str): element to replace
            new_path_element (str): element to replace with

        """
        self.logger.debug("Replacing PATH element {0} with {1}".format(old_path_element, new_path_element))
        self._internal_set_path([x if x != old_path_element else new_path_element for x in self.active_path])

    def replace_pypath_element(self, old_pypath_element, new_pypath_element):
        """Replaces the PYPATH element.

        Generates a new PYPATH by iterating through the old PYPATH and replacing
        old_pypath_element with new_pypath_element where it is found.

        Args:
            old_pypath_element (str): element to replace
            new_pypath_element (str): element to replace with

        """
        self.logger.debug("Replacing PYPATH element {0} with {1}".format(old_pypath_element, new_pypath_element))
        self._internal_set_pypath([x if x != old_pypath_element else new_pypath_element for x in self.active_pypath])

    def remove_path_element(self, path_element):
        """Removes the PATH element.

        Generates a new PATH by iterating through the old PATH and removing
        path_element if it is found.

        Args:
            path_element (str): path element to remove
        """
        self.logger.debug("Removing PATH element {0}".format(path_element))
        self._internal_set_path([x for x in self.active_path if x != path_element])

    def remove_pypath_element(self, pypath_element):
        """Removes the PYPATH element.

        Generates a new PYPATH by iterating through the old PYPATH and removing
        pypath_element if it is found.

        Args:
            pypath_element (str): pypath element to remove
        """
        self.logger.debug("Removing PYPATH element {0}".format(pypath_element))
        self._internal_set_pypath([x for x in self.active_pypath if x != pypath_element])

    def get_build_var(self, var_name):
        """Gets the build variable.

        Args:
            var_name (str): variable to get the value of
        Returns:
            (obj): value associated with the var_name
        """
        return self.active_buildvars.GetValue(var_name)

    def set_build_var(self, var_name, var_data):
        """Sets the build var.

        Args:
            var_name (str): variable to set the value for
            var_data (obj): data to set

        WARNING: Unlike `set_shell_var`, this only sets the variable in the
        `VarDict`
        """
        self.logger.debug(
            "Updating BUILD VAR element '%s': '%s'." % (var_name, var_data))
        self.active_buildvars.SetValue(var_name, var_data, '', overridable=True)

    def get_shell_var(self, var_name):
        """Gets the shell variable.

        Args:
            var_name (str): variable to get the value of

        Returns:
            (obj): value associated with the var name
        """
        return self.active_environ.get(var_name, None)

    def set_shell_var(self, var_name, var_data):
        """Sets the shell variable.

        Args:
            var_name (str): variable to set the value for
            var_data (obj): data to set

        The variable is set both in the `VarDict` and in the OS
        """
        # Check for the "special" shell vars.
        if var_name.upper() == 'PATH':
            self.set_path(var_data)
        elif var_name.upper() == 'PYTHONPATH':
            self.set_pypath(var_data)
        else:
            self.logger.debug(
                "Updating SHELL VAR element '%s': '%s'." % (var_name, var_data))
            self.active_environ[var_name] = var_data
            os.environ[var_name] = var_data


def GetEnvironment():
    """Returns the environment.

    Returns:
        (ShellEnvironment): Singleton class
    """
    return ShellEnvironment()


def GetBuildVars():
    """The current checkpoint buildvar values.

    Returns:
        (VarDict): A special dictionary containing build vars
    """
    #
    # Tricky!
    # Define a wrapper class that always forwards commands to the
    # BuildVars associated with the current environment.
    #
    # Will be deprecated.
    #
    class BuildVarsWrapper(object):
        def __init__(self):
            self.internal_shell_env = ShellEnvironment()

        def __getattr__(self, attrname):
            # Instead, invoke on the active BuildVars object.
            return getattr(self.internal_shell_env.active_buildvars, attrname)

    return BuildVarsWrapper()


#
# TODO: These are convenience methods that should be deprecated.
#
checkpoint_list = list()


def CheckpointBuildVars():
    """Creates a checkpoint [a screenshot in time] of all current build var values."""
    global checkpoint_list
    new_checkpoint = ShellEnvironment().checkpoint()
    checkpoint_list.append(new_checkpoint)
    MY_LOGGER.debug("Created checkpoint {0} for build vars".format(new_checkpoint))


def RevertBuildVars():
    """Reverts all build var values to the most recent checkpoint."""
    global checkpoint_list
    if len(checkpoint_list) > 0:
        last_checkpoint = checkpoint_list.pop()
        MY_LOGGER.debug("Reverting to checkpoint {0} for build vars".format(last_checkpoint))
        ShellEnvironment().restore_checkpoint(last_checkpoint)
    else:
        MY_LOGGER.getLogger("No more checkpoints!")
        raise RuntimeError("No more checkpoints!")



# @file edk2_ci_setup.py
# Resolves all dependent repos for a CI environment.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code that supports CI/CD via the ci_setup invocable.

Contains a CISetupSettingsManager that must be subclassed in a build settings
file. This provides platform specific information to Edk2CiSetup invocable
while allowing the invocable itself to remain platform agnostic.
"""
import os
import logging

from edk2toolext.invocables.edk2_multipkg_aware_invocable import Edk2MultiPkgAwareInvocable
from edk2toolext.invocables.edk2_multipkg_aware_invocable import MultiPkgAwareSettingsInterface
from edk2toolext.environment import repo_resolver


class CiSetupSettingsManager(MultiPkgAwareSettingsInterface):
    """Platform specific settings for Edk2CiSetup.

    Provide information necessary for `stuart_ci_setup.exe` or
    `edk2_ci_setup.py` to successfully execute.

    Example: Example: Overriding CiSetupSettingsManager
        ```python
        from edk2toolext.invocables.edk2_ci_setup import CiSetupSettingsManager
        class CiManager(CiSetupSettingsManager):
            def GetDependencies(self):
                return {
                    "Path": "/Common/MU",
                    "Url":  "https://github.com/Microsoft/mu_tiano_plus.git"
                }
        ```
    """

    def GetDependencies(self):
        """Get any Git Repository Dependendencies.

        This list of repositories will be resolved during the setup step.

        TIP: Optional Override in subclass

        TIP: Return an iterable of dictionary objects with the following fields
        {
            Path: <required> Workspace relative path
            Url: <required> Url of git repo
            Commit: <optional> Commit to checkout of repo
            Branch: <optional> Branch to checkout (will checkout most recent commit in branch)
            Full: <optional> Boolean to do shallow or Full checkout.  (default is False)
            ReferencePath: <optional> Workspace relative path to git repo to use as "reference"
        }
        """
        return []


class Edk2CiBuildSetup(Edk2MultiPkgAwareInvocable):
    """Invocable supporting an iterative multi-package build and test process leveraging CI build plugins.

    Edk2CiBuildSetup sets up the necessary environment for Edk2CiBuild by preparing all necessary submodules.
    """
    def AddCommandLineOptions(self, parser):
        """Adds command line arguments to Edk2CiBuild."""
        parser.add_argument('-ignore', '--ignore-git', dest="git_ignore", action="store_true",
                            help="Whether to ignore errors in the git cloning process", default=False)
        parser.add_argument('--omnicache', '--reference', dest='omnicache_path',
                            default=os.environ.get('OMNICACHE_PATH'))
        parser.add_argument('-force', '--force-git', dest="git_force", action="store_true",
                            help="Whether to force git repos to clone in the git cloning process", default=False)
        parser.add_argument('-update-git', '--update-git', dest="git_update", action="store_true",
                            help="Whether to update git repos as needed in the git cloning process", default=False)
        super().AddCommandLineOptions(parser)

    def RetrieveCommandLineOptions(self, args):
        """Retrieve command line options from the argparser."""
        self.git_ignore = args.git_ignore
        self.git_force = args.git_force
        self.git_update = args.git_update
        self.omnicache_path = args.omnicache_path
        if (self.omnicache_path is not None) and (not os.path.exists(self.omnicache_path)):
            logging.warning(f"Omnicache path set to invalid path: {args.omnicache_path}")
            self.omnicache_path = None

        super().RetrieveCommandLineOptions(args)

    def GetVerifyCheckRequired(self):
        """Will not verify environment."""
        return False

    def GetSettingsClass(self):
        """Returns the CiSetupSettingsManager class.

        WARNING: CiSetupSettingsManager must be subclassed in your platform settings file.
        """
        return CiSetupSettingsManager

    def GetLoggingFileName(self, loggerType):
        """Returns the filename (CISETUP) of where the logs for the Edk2CiBuild invocable are stored in."""
        return "CISETUP"

    def Go(self):
        """Executes the core functionality of the Edk2CiSetup invocable."""
        setup_dependencies = self.PlatformSettings.GetDependencies()
        logging.debug(f"Dependencies list {setup_dependencies}")
        repos = repo_resolver.resolve_all(self.GetWorkspaceRoot(),
                                          setup_dependencies,
                                          ignore=self.git_ignore, force=self.git_force,
                                          update_ok=self.git_update, omnicache_dir=self.omnicache_path)

        logging.info(f"Repo resolver resolved {repos}")

        return 0 if None not in repos else -1


def main():
    """Entry point invoke Edk2CiBuild."""
    Edk2CiBuildSetup().Invoke()



# @file NugetPublishing.py
# This tool allows a user to create a configuration for nuget as well as
# pack and push (publishing) a release to a feed.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Provides configuration, packing, and publishing nuget packages to a release feed."""
import os
import sys
import argparse
import logging
import yaml
import xml.etree.ElementTree as etree
import shutil
import datetime
from io import StringIO
from edk2toolext.environment.extdeptypes.nuget_dependency import NugetDependency
from edk2toollib.utility_functions import RunCmd

OPEN_SOURCE_INITIATIVE_URL = "https://opensource.org/licenses/"
LICENSE_TYPE_SUPPORTED = {
    "BSD2": OPEN_SOURCE_INITIATIVE_URL + "BSD-2-Clause",
    "BSD3": OPEN_SOURCE_INITIATIVE_URL + "BSD-3-Clause",
    "APACHE2": OPEN_SOURCE_INITIATIVE_URL + "Apache-2.0",
    "MSPL": OPEN_SOURCE_INITIATIVE_URL + "MS-PL",  # Microsoft Public License
    "MIT": OPEN_SOURCE_INITIATIVE_URL + "MIT",
    "BSDpP": OPEN_SOURCE_INITIATIVE_URL + "BSDplusPatent",  # BSD + Patent
}


class NugetSupport(object):
    """Support object for Nuget Publishing tool to configure NuPkg information, pack and send."""
    # NOTE: This *should* have a namespace (http://schemas.microsoft.com/packaging/2010/07/nuspec.xsd)
    #       but ElementTree is incredibly stupid with namespaces.
    NUSPEC_TEMPLATE_XML = r'''<?xml version="1.0" encoding="utf-8"?>
<package>
    <metadata>
        <!-- Required elements-->
        <id></id>
        <version></version>
        <description></description>
        <authors></authors>

        <!-- Optional elements -->
        <licenseUrl></licenseUrl>
        <releaseNotes></releaseNotes>
        <projectUrl></projectUrl>
        <copyright></copyright>
        <tags></tags>
    </metadata>
    <!-- Optional 'files' node -->
    <files>
        <file src="" target="" />
    </files>
</package>'''

    RELEASE_NOTE_SHORT_STRING_MAX_LENGTH = 500

    def __init__(self, Name=None, ConfigFile=None):
        """Inits a new NugetSupport object.

        for new instances without existing config provide the Name parameter.
        for creating instance based on config file provide the path to the ConfigFile
        """
        self.Name = Name
        self.TempFileToDelete = []  # everytime a temp is created add to list to cleanup
        self.NewVersion = None
        self.ConfigChanged = False

        if (ConfigFile is not None):
            self.FromConfigfile(ConfigFile)
            self.Name = self.ConfigData["name"]
        else:
            if (Name is None):
                raise ValueError("Cannot construct object with both Name and ConfigFile as None")
            self.ConfigData = {"name": Name}
            self.Config = None

    def CleanUp(self):
        """Remove all temporary files."""
        logging.debug("CleanUp Called.  Deleting all Temp Files")
        for a in self.TempFileToDelete:
            os.remove(a)

    def ToConfigFile(self, filepath=None):
        """Save config to a yaml file."""
        if (not self.ConfigChanged):
            logging.debug("No Config Changes.  Skip Writing config file")
            return 0

        if (filepath is None and self.Config is None):
            logging.error("No Config File to save to.")
            return -1

        if (filepath is not None):
            self.Config = filepath

        if (filepath is None):
            logging.error("No filepath for Config File")

        with open(filepath, "w") as c:
            yaml.dump(self.ConfigData, c, indent=4)
        logging.debug("Wrote config file to: %s" % filepath)
        self.ConfigChanged = False
        return 0

    def FromConfigfile(self, filepath):
        """Load config from a yaml file."""
        self.Config = filepath
        with open(self.Config, "r") as c:
            self.ConfigData = yaml.safe_load(c)

    def SetBasicData(self, authors, license, project, description, server, copyright):
        """Set basic data in the config data."""
        self.ConfigData["author_string"] = authors
        self.ConfigData["license_url"] = license
        self.ConfigData["project_url"] = project
        self.ConfigData["description_string"] = description
        self.ConfigData["server_url"] = server

        if not copyright:
            copyright = "Copyright %d" % datetime.date.today().year
        self.ConfigData["copyright_string"] = copyright

        self.ConfigData["tags_string"] = ""

        self.ConfigChanged = True

    def UpdateCopyright(self, copyright):
        """Update copyright in the config data."""
        self.ConfigData["copyright_string"] = copyright
        self.ConfigChanged = True

    def UpdateTags(self, tags=[]):
        """Update tags in the config data."""
        self.ConfigData["tags_string"] = " ".join(tags)
        self.ConfigChanged = True

    def Print(self):
        """Print info about the Nuget Object."""
        print("=======================================")
        print(" Name:        " + self.Name)
        if (self.Config):
            print(" ConfigFile:  " + self.Config)
        else:
            print(" ConfigFile:  NOT SET")

        for k, v in self.ConfigData.items():
            print(" %s:   %s" % (k, v))

        print("----------------------------------------")
        print(" Temp Files List: ")
        for a in self.TempFileToDelete:
            print("    " + a)
        print("-----------------------------------------")
        print("=======================================")

    def LogObject(self):
        """Logs info about Nuget Object to the logger."""
        logging.debug("=======================================")
        logging.debug(" Name:        " + self.Name)
        if (self.Config):
            logging.debug(" ConfigFile:  " + self.Config)
        else:
            logging.debug(" ConfigFile:  NOT SET")

        for k, v in self.ConfigData.items():
            logging.debug(" %s:   %s" % (k, v))

        logging.debug("----------------------------------------")
        logging.debug(" Temp Files List: ")
        for a in self.TempFileToDelete:
            logging.debug("    " + a)
        logging.debug("-----------------------------------------")
        logging.debug("=======================================")

    #
    # given NugetSupport object
    # create a nuspec file for packing
    #

    def _MakeNuspecXml(self, ContentDir, ReleaseNotesText=None):
        package = etree.fromstring(NugetSupport.NUSPEC_TEMPLATE_XML)
        meta = package.find("./metadata")
        meta.find("id").text = self.Name
        meta.find("version").text = self.NewVersion
        meta.find("authors").text = self.ConfigData["author_string"]
        meta.find("licenseUrl").text = self.ConfigData["license_url"]
        meta.find("projectUrl").text = self.ConfigData["project_url"]
        meta.find("description").text = self.ConfigData["description_string"]
        meta.find("copyright").text = self.ConfigData["copyright_string"]
        if "tags_string" in self.ConfigData:
            meta.find("tags").text = self.ConfigData["tags_string"]
        files = package.find("files")
        f = files.find("file")
        f.set("target", self.Name)
        f.set("src", ContentDir + "\\**\\*")

        if (ReleaseNotesText is not None):
            logging.debug("Make Nuspec Xml - ReleaseNotesText is not none.")
            #
            # Make sure it doesn't exceed reasonable length of string
            #
            if (len(ReleaseNotesText) > NugetSupport.RELEASE_NOTE_SHORT_STRING_MAX_LENGTH):
                logging.info("Make Nuspec Xml - ReleaseNotesText too long.  Length is (%d)" % len(ReleaseNotesText))
                logging.debug("Original ReleaseNotesText is: %s" % ReleaseNotesText)
                # cut it off at max length
                ReleaseNotesText = ReleaseNotesText[:NugetSupport.RELEASE_NOTE_SHORT_STRING_MAX_LENGTH]
                # walk back to trim at last end of sentence
                ReleaseNotesText = ReleaseNotesText.rpartition(".")[0].strip()
                logging.debug("New ReleaseNotesText is: %s" % ReleaseNotesText)

            meta.find("releaseNotes").text = ReleaseNotesText
        else:
            logging.debug("Make Nuspec Xml - ReleaseNotesText None. Removing element from nuspec.")
            meta.remove(meta.find("releaseNotes"))

        return etree.tostring(package)

    def _GetNuPkgFileName(self, version):
        # Nuget removes leading zeros so to match we must do the same
        s = self.Name + "."
        append_tag = None
        parts = version.split(".")
        if "-" in parts[-1]:
            parts[-1], append_tag = parts[-1].split("-")
        int_parts = [str(int(a)) for a in parts]

        # nuget must have at least x.y.z and will make zero any element undefined
        for _ in range(len(int_parts), 3):
            int_parts.append("0")
        # Join the integers together
        s += ".".join(int_parts)
        if append_tag is not None:
            s += f"-{append_tag}"
        s += ".nupkg"
        return s

    def Pack(self, version, OutputDirectory, ContentDir, RelNotesText=None):
        """Pack the current contents into Nupkg."""
        self.NewVersion = version

        # content must be absolute path in nuspec otherwise it is assumed
        # relative to nuspec file.
        cdir = os.path.abspath(ContentDir)

        # make nuspec file
        xmlstring = self._MakeNuspecXml(cdir, RelNotesText)
        nuspec = os.path.join(OutputDirectory, self.Name + ".nuspec")
        self.TempFileToDelete.append(nuspec)
        f = open(nuspec, "wb")
        f.write(xmlstring)
        f.close()

        # run nuget
        cmd = NugetDependency.GetNugetCmd()
        cmd += ["pack", nuspec]
        cmd += ["-OutputDirectory", '"' + OutputDirectory + '"']
        cmd += ["-Verbosity", "detailed"]
        # cmd += ["-NonInteractive"]
        ret = RunCmd(cmd[0], " ".join(cmd[1:]))

        if (ret != 0):
            logging.error("Failed on nuget commend.  RC = 0x%x" % ret)
            return ret

        self.NuPackageFile = os.path.join(OutputDirectory, self._GetNuPkgFileName(self.NewVersion))
        self.TempFileToDelete.append(self.NuPackageFile)
        return ret

    def Push(self, nuPackage, apikey):
        """Push nuget package to the server.

        Raises:
            (Exception): file path is invalid
        """
        if (not os.path.isfile(nuPackage)):
            raise Exception("Invalid file path for NuPkg file")
        logging.debug("Pushing %s file to server %s" % (nuPackage, self.ConfigData["server_url"]))

        cmd = NugetDependency.GetNugetCmd()
        cmd += ["push", nuPackage]
        cmd += ["-Verbosity", "detailed"]
        # cmd += ["-NonInteractive"]
        cmd += ["-Source", self.ConfigData["server_url"]]
        cmd += ["-ApiKey", apikey]
        output_buffer = StringIO()
        ret = RunCmd(cmd[0], " ".join(cmd[1:]), outstream=output_buffer)

        if (ret != 0):
            # Rewind the buffer and capture the contents.
            output_buffer.seek(0)
            output_contents = output_buffer.read()

            # Check for the API message.
            if "API key is invalid".lower() in output_contents.lower():
                logging.critical("API key is invalid. Please use --ApiKey to provide a valid key.")

            # Generic error.
            logging.error("Failed on nuget commend.  RC = 0x%x" % ret)

        return ret


def GatherArguments():
    """Adds CLI arguments for controlling the nuget_publishing tool."""
    tempparser = argparse.ArgumentParser(
        description='Nuget Helper Script for creating, packing, and pushing packages', add_help=False)
    tempparser.add_argument('--Operation', dest="op", choices=["New", "Pack", "Push", "PackAndPush"], required=True)

    # Get the operation the user wants to do
    (args, rest) = tempparser.parse_known_args()

    # now build up the real parser with required parameters
    parser = argparse.ArgumentParser(description='Nuget Helper Script for creating, packing, and pushing packages')
    parser.add_argument("--Dirty", dest="Dirty", action="store_true", help="Keep all temp files", default=False)
    parser.add_argument('--Operation', dest="Operation", choices=["New", "Pack", "Push", "PackAndPush"], required=True)
    parser.add_argument("--OutputLog", dest="OutputLog", help="Create an output log file")

    if (args.op.lower() == "new"):
        parser.add_argument("--ConfigFileFolderPath", dest="ConfigFileFolderPath",
                            help="<Required>Path to folder to save new config file to", required=True)
        parser.add_argument('--Name',
                            dest='Name',
                            help='<Required> The unique id/name of the package.  This is a string naming the package',
                            required=True)
        parser.add_argument('--Author', dest="Author", help="<Required> Author string for publishing", required=True)
        parser.add_argument("--ProjectUrl", dest="Project", help="<Required> Project Url", required=True)
        g = parser.add_mutually_exclusive_group(required=True)
        g.add_argument('--CustomLicenseUrl', dest="LicenseUrl",
                       help="<Optional> http url for custom license file.  Can use LicenseType for standard licenses")
        g.add_argument('--LicenseType', dest="LicenseType",
                       choices=LICENSE_TYPE_SUPPORTED.keys(), help="Standard Licenses")
        parser.add_argument('--Description', dest="Description",
                            help="<Required> Description of package.", required=True)
        parser.add_argument("--FeedUrl", dest="FeedUrl",
                            help="<Required>Feed Url of the nuget server feed", required=True)
        parser.add_argument('--Copyright', dest="Copyright", help="Copyright string", required=False)

    elif (args.op.lower() == "pack" or args.op.lower() == "packandpush"):
        parser.add_argument("--ConfigFilePath", dest="ConfigFilePath",
                            help="<Required>Path to config file", required=True)
        parser.add_argument('--Version', dest="Version", help="<Required> Version to publish", required=True)
        parser.add_argument('--ReleaseNotesText', dest="ReleaseNotes",
                            help="<Optional>Release Notes String", required=False)
        parser.add_argument('--InputFolderPath', dest="InputFolderPath",
                            help="<Required>Relative/Absolute Path to folder containing content to pack.",
                            required=True)
        parser.add_argument('--Copyright', dest="Copyright", help="<Optional>Change the Copyright string")
        parser.add_argument('--t', "-tag", dest="Tags", type=str,
                            help="<Optional>Add tags to the nuspec. Multiple are --t Tag1,Tag2 or --t Tag1 --t Tag2",
                            action="append", default=[])
        parser.add_argument('--ApiKey', dest="ApiKey",
                            help="<Optional>Api key to use. Default is 'VSTS' which will invoke interactive login",
                            default="VSTS")

    elif (args.op.lower() == "push"):
        parser.add_argument("--ConfigFilePath", dest="ConfigFilePath",
                            help="<Required>Path to config file",
                            required=True)
        parser.add_argument('--PackageFile', dest="PackageFile", help="<Required>Path To Package File", required=True)
        parser.add_argument('--ApiKey', dest="ApiKey",
                            help="<Optional>Api key to use. Default is 'VSTS' which will invoke interactive login",
                            default="VSTS")

    if (args.op.lower() == "pack"):
        parser.add_argument('--OutputFolderPath',
                            dest="OutputFolderPath",
                            help="<Optional>Output folder where nupkg will be saved.  Default is cwd",
                            default=os.getcwd())

    return parser.parse_args()


def main():
    """Entry point into nuget_publishing after initial configuration."""
    args = GatherArguments()
    ret = 0

    # setup file based logging if outputReport specified
    if (args.OutputLog):
        if (len(args.OutputLog) < 2):
            logging.critical("the output log file parameter is invalid")
            return -2

        # setup file based logging
        filelogger = logging.FileHandler(filename=args.OutputLog, mode='w')
        filelogger.setLevel(logging.DEBUG)
        logging.getLogger('').addHandler(filelogger)

    logging.info("Log Started: " + datetime.datetime.strftime(datetime.datetime.now(), "%A, %B %d, %Y %I:%M%p"))

    TempOutDir = None
    NuPkgFilePath = None

    if (args.Operation.lower() == "new"):
        logging.critical("Generating new nuget configuration...")
        logging.debug("Checking input parameters for new")
        ConfigFilePath = os.path.join(args.ConfigFileFolderPath, args.Name.strip() + ".config.yaml")

        if (not os.path.isdir(args.ConfigFileFolderPath)):
            logging.critical("Config File Folder Path doesn't exist.  %s" % args.ConfigFileFolderPath)
            raise Exception("Invalid Config File Folder.  Doesn't exist")

        if (os.path.isfile(ConfigFilePath)):
            logging.critical("Config File already exists at that path.  %s" % ConfigFilePath)
            raise Exception("Can't Create New Config file when file already exists")

        nu = NugetSupport(Name=args.Name)

        # license
        license_url = args.LicenseUrl
        if (args.LicenseType is not None):
            license_url = LICENSE_TYPE_SUPPORTED[args.LicenseType]
        nu.SetBasicData(args.Author, license_url, args.Project, args.Description, args.FeedUrl, args.Copyright)
        nu.LogObject()
        ret = nu.ToConfigFile(ConfigFilePath)
        return ret

    elif (args.Operation.lower() == "pack" or args.Operation.lower() == "packandpush"):
        logging.critical("Creating nuget package")
        logging.debug("Checking input parameters for packing")
        # check args
        if (not os.path.isfile(args.ConfigFilePath)):
            logging.critical("Invalid Config File (%s).  File doesn't exist" % args.ConfigFilePath)
            raise Exception("Invalid Config File.  File doesn't exist")
        if (not os.path.isdir(args.InputFolderPath)):
            logging.critical("Invalid Input folder (%s).  Folder doesn't exist" % args.InputFolderPath)
            raise Exception("Invalid Input folder.  folder doesn't exist")
        contents = os.listdir(args.InputFolderPath)
        logging.debug("Input Folder contains %d files" % len(contents))
        if (len(contents) == 0):
            logging.critical("No binary contents to pack in %s" % args.InputFolderPath)
            raise Exception("No binary contents to package")

        # make a temp dir for the pack operation which actually creates files
        TempOutDir = os.path.join(os.getcwd(), "_TEMP_" + str(datetime.datetime.now().time()).replace(":", "_"))
        os.mkdir(TempOutDir)

        nu = NugetSupport(ConfigFile=args.ConfigFilePath)
        if (args.Copyright is not None):
            nu.UpdateCopyright(args.Copyright)
        if (len(args.Tags) > 0):
            tagListSet = set()
            for item in args.Tags:  # Parse out the individual packages
                item_list = item.split(",")
                for individual_item in item_list:
                    # in case cmd line caller used Windows folder slashes
                    individual_item = individual_item.replace("\\", "/")
                    tagListSet.add(individual_item.strip())
            tagList = list(tagListSet)
            nu.UpdateTags(tagList)
        '''
        ret = nu.ToConfigFile()
        if (ret != 0):
            logging.error("Failed to save config file.  Return Code 0x%x" % ret)
            return ret
        '''

        ret = nu.Pack(args.Version, TempOutDir, args.InputFolderPath, args.ReleaseNotes)
        if (ret != 0):
            logging.error("Failed to pack.  Return Code 0x%x" % ret)
            return ret

        NuPkgFilePath = nu.NuPackageFile

    if (args.Operation.lower() == "pack"):
        if (not os.path.isdir(args.OutputFolderPath)):
            logging.critical("Invalid Pack Output Folder (%s).  Folder doesn't exist" % args.OutputFolderPath)
            raise Exception("Invalid Output folder.  folder doesn't exist")
        # since it is pack only lets copy nupkg file to output
        shutil.copyfile(NuPkgFilePath, os.path.join(args.OutputFolderPath, os.path.basename(NuPkgFilePath)))
        NuPkgFilePath = os.path.join(args.OutputFolderPath, os.path.basename(NuPkgFilePath))

    if (args.Operation.lower() == "push"):
        # set the parameters for push
        logging.debug("Checking input parameters for push")
        # check args
        if (not os.path.isfile(args.ConfigFilePath)):
            logging.critical("Invalid Config File (%s).  File doesn't exist" % args.ConfigFilePath)
            raise Exception("Invalid Config File.  File doesn't exist")
        NuPkgFilePath = args.PackageFile
        nu = NugetSupport(ConfigFile=args.ConfigFilePath)

    if (args.Operation.lower() == "push" or args.Operation.lower() == "packandpush"):
        # do the pushing
        logging.critical("Pushing the package")
        logging.debug("NuPkgFilePath is %s" % NuPkgFilePath)
        # check args
        if (not os.path.isfile(NuPkgFilePath)):
            logging.critical("NuPkgFilePath is not valid file.  %s" % NuPkgFilePath)
            raise Exception("Invalid Pkg File.  File doesn't exist")
        ret = nu.Push(NuPkgFilePath, args.ApiKey)

    nu.LogObject()
    nu.ToConfigFile(args.ConfigFilePath)  # save any changes
    if (not args.Dirty):
        nu.CleanUp()
        if (TempOutDir is not None):
            os.removedirs(TempOutDir)
    return ret


def go():
    """Main entry into the nuget publishing tool."""
    # setup main console as logger
    logger = logging.getLogger('')
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(levelname)s - %(message)s")
    console = logging.StreamHandler()
    console.setLevel(logging.WARNING)
    console.setFormatter(formatter)
    logger.addHandler(console)

    # call main worker function
    retcode = main()

    if retcode != 0:
        logging.critical("Failed.  Return Code: %d" % retcode)
    else:
        logging.critical("Success!")
    # end logging
    logging.shutdown()
    sys.exit(retcode)


if __name__ == '__main__':
    go()



## @file setup.py
# This contains setup info for edk2-pytool-extensions pip module
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Setup info for edk2-pytool-extensions pip module."""
import setuptools
from setuptools.command.sdist import sdist
from setuptools.command.install import install
from setuptools.command.develop import develop
from edk2toolext.bin.nuget import DownloadNuget

with open("readme.md", "r") as fh:
    long_description = fh.read()


class PostSdistCommand(sdist):
    """Post-sdist."""
    def run(self): # noqa
        # we need to download nuget so throw the exception if we don't get it
        DownloadNuget()
        sdist.run(self)


class PostInstallCommand(install):
    """Post-install."""
    def run(self): # noqa
        try:
            DownloadNuget()
        except Exception:
            pass
        install.run(self)


class PostDevCommand(develop):
    """Post-develop."""
    def run(self): # noqa
        try:
            DownloadNuget()
        except Exception:
            pass
        develop.run(self)


setuptools.setup(
    name="edk2-pytool-extensions",
    author="Tianocore Edk2-PyTool-Extensions team",
    author_email="sean.brogan@microsoft.com",
    description="Python tools supporting UEFI EDK2 firmware development",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tianocore/edk2-pytool-extensions",
    license='BSD-2-Clause-Patent',
    packages=setuptools.find_packages(),
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    cmdclass={
        'sdist': PostSdistCommand,
        'install': PostInstallCommand,
        'develop': PostDevCommand,
    },
    include_package_data=True,
    entry_points={
        'console_scripts': ['stuart_setup=edk2toolext.invocables.edk2_setup:main',
                            'stuart_update=edk2toolext.invocables.edk2_update:main',
                            'stuart_build=edk2toolext.invocables.edk2_platform_build:main',
                            'stuart_ci_build=edk2toolext.invocables.edk2_ci_build:main',
                            'stuart_ci_setup=edk2toolext.invocables.edk2_ci_setup:main',
                            'stuart_pr_eval=edk2toolext.invocables.edk2_pr_eval:main',
                            'omnicache=edk2toolext.omnicache:main',
                            'nuget-publish=edk2toolext.nuget_publishing:go',
                            'sig_db_tool=edk2toolext.uefi.sig_db_tool:main',
                            'firmware_policy_tool=edk2toolext.windows.policy.firmware_policy_tool:main',
                            'edk2_capsule_tool=edk2toolext.capsule.capsule_tool:main',
                            'versioninfo_tool=edk2toolext.versioninfo.versioninfo_tool:main',
                            'validate_image_tool=edk2toolext.image_validation:main']
    },
    install_requires=[
        'pyyaml>=5.3.1',
        'edk2-pytool-library>=0.12.1',
        'pefile>=2019.4.18',
        'semantic_version>=2.10.0'
    ],
    extras_require={
        'openssl': ['pyopenssl']
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers"
    ]
)



# @file self_describing_environment.py
# This module contains code that is capable of scanning the source tree for
# files that describe the source and dependencies and acting upon those files.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""An environment capable of scanning the source tree.

Scans the environment for files that describe the source and dependencies
and then acts upon those files.
"""
import os
import logging
import pathlib
from edk2toolext.environment import shell_environment
from edk2toolext.environment import environment_descriptor_files as EDF
from edk2toolext.environment import external_dependency
from multiprocessing import dummy
import time


ENVIRONMENT_BOOTSTRAP_COMPLETE = False
ENV_STATE = None


class self_describing_environment(object):
    """An environment capable of scanning the source tree.

    Scans the environment for files that describe the source and dependencies
    and then acts upon those files.
    """
    def __init__(self, workspace_path, scopes=(), skipped_dirs=()):
        """Inits an empty self describing environment."""
        logging.debug("--- self_describing_environment.__init__()")
        logging.debug(f"Skipped directories specified = {skipped_dirs}")
        super(self_describing_environment, self).__init__()

        self.workspace = workspace_path

        # Determine the final set of scopes.
        # Start with the provided set.
        self.scopes = scopes

        # Allow certain directories to be skipped
        self.skipped_dirs = tuple(map(pathlib.Path, (os.path.join(self.workspace, d) for d in skipped_dirs)))

        # Validate that all scopes are unique.
        if len(self.scopes) != len(set(self.scopes)):
            raise ValueError(f"All scopes must be unique!\n\t{self.scopes}")

        self.paths = None
        self.extdeps = None
        self.plugins = None

    def _gather_env_files(self, ext_strings, base_path):
        logging.debug("--- self_describing_environment._gather_env_files()")
        # Make sure that the search extension matches easily.
        search_files = tuple(ext_string.lower() for ext_string in ext_strings)

        # Walk all of the directories under base_path and find all files
        # matching the extension.
        matches = {}
        for root, dirs, files in os.walk(base_path, topdown=True):
            # Check to see whether any of these directories should be skipped..
            dirs[:] = [d
                       for d
                       in dirs
                       if pathlib.Path(root, d) not in self.skipped_dirs
                       and pathlib.Path(root, d).name != '.git']

            # Check for any files that match the extensions we're looking for.
            for file in files:
                for search_file in search_files:
                    if file.lower().endswith(search_file + ".json") or file.lower().endswith(search_file + ".yaml"):
                        if search_file in matches:
                            matches[search_file].append(
                                os.path.join(root, file))
                        else:
                            matches[search_file] = [os.path.join(root, file)]

        return matches

    def load_workspace(self):
        """Loads the workspace."""
        logging.debug("--- self_describing_environment.load_workspace()")
        logging.debug("Loading workspace: %s" % self.workspace)
        logging.debug("  Including scopes: %s" % ', '.join(self.scopes))

        # First, we need to get all of the files that describe our environment.
        env_files = self._gather_env_files(('path_env', 'ext_dep', 'plug_in'), self.workspace)

        # Next, get a list of all our scopes
        all_scopes_lower = [x.lower() for x in self.scopes]

        # Now that the files have been found, load them, sort them, and filter them
        # so they can be applied to the environment.

        # We need to convert them from files to descriptors
        all_descriptors = list()

        # helper function to get all the descriptors of a type and cast them
        def _get_all_descriptors_of_type(key, class_type):
            if key not in env_files:
                return tuple()
            return tuple(class_type(desc_file) for desc_file in env_files[key])

        # Collect all the descriptors of each type
        all_descriptors.extend(_get_all_descriptors_of_type('path_env', EDF.PathEnvDescriptor))
        all_descriptors.extend(_get_all_descriptors_of_type('ext_dep', EDF.ExternDepDescriptor))
        all_descriptors.extend(_get_all_descriptors_of_type('plug_in', EDF.PluginDescriptor))

        # Get the properly scoped descriptors by checking if the scope is in the list of all the scopes
        scoped_desc_gen = [x for x in all_descriptors if x.descriptor_contents['scope'].lower() in all_scopes_lower]
        scoped_descriptors = list(scoped_desc_gen)

        # Check that each found item has a unique ID, that's an error if it isn't
        allids_gen = [x.descriptor_contents['id'].lower() for x in scoped_descriptors if 'id' in x.descriptor_contents]
        all_ids = list(allids_gen)
        all_unique_ids = set(all_ids)
        if len(all_ids) != len(all_unique_ids):
            logging.error("Multiple descriptor files share the same id")
            all_unique_id_dict = {}
            for desc_id in all_ids:
                dict_id_seen = desc_id not in all_unique_id_dict
                all_unique_id_dict[desc_id] = 1 if dict_id_seen else all_unique_id_dict[desc_id] + 1
            for desc_id in all_unique_id_dict:
                if all_unique_id_dict[desc_id] == 1:
                    continue
                # get the descriptors
                desc_of_id = [
                    x for x in scoped_descriptors
                    if 'id' in x.descriptor_contents and x.descriptor_contents['id'].lower() == desc_id
                ]
                paths_of_desc_of_id = [x.file_path for x in desc_of_id]
                invalid_desc_paths = f"{os.pathsep} ".join(paths_of_desc_of_id)
                logging.error(f"Descriptors that have this id {desc_id}: {invalid_desc_paths}")
            raise RuntimeError("Multiple descriptor files share the same id")

        # Now check for overrides, first get a list of all the descriptors that have an override id tag
        override_descriptors = [x for x in scoped_descriptors if "override_id" in x.descriptor_contents]
        active_overrides = {}
        for desc in override_descriptors:
            override_id = desc.descriptor_contents["override_id"].lower()
            # we found this was already overriden, make sure to let the user know
            if override_id in active_overrides:
                logging.warning("A descriptor file is trying to override a file that was previously overriden")
                logging.warning(f"File ID being overriden: {override_id}")
                logging.warning(f"Previous override: {active_overrides[override_id].file_path}")
                logging.warning(f"New override: {desc.file_path}")
                raise RuntimeError(f"Multiple descriptor files share the same override_id: {override_id}")
            active_overrides[override_id] = desc

        # Now we filter the overriden id's out and debug to the user whether we are including them or not
        overriden_ids = active_overrides.keys()
        final_descriptors = []
        for desc in scoped_descriptors:
            desc_file = desc.file_path
            if 'id' in desc.descriptor_contents:
                desc_id = desc.descriptor_contents['id'].lower()
                if desc_id in overriden_ids:
                    override = active_overrides[desc_id]
                    desc_name = f"{desc_file}:{desc_id}"
                    override_name = f"{override.file_path}"
                    logging.debug(f"Skipping descriptor {desc_name} as it is being overridden by {override_name}.")
                    continue
            # add them to the final list
            desc_scope = desc.descriptor_contents['scope']
            logging.debug(f"Adding descriptor {desc_file} to the environment with scope {desc_scope}")
            final_descriptors.append(desc)

        # Finally, sort them back in the right categories
        self.paths = list([x.descriptor_contents for x in final_descriptors if isinstance(x, EDF.PathEnvDescriptor)])
        self.extdeps = list(
            [x.descriptor_contents for x in final_descriptors if isinstance(x, EDF.ExternDepDescriptor)])
        self.plugins = list([x.descriptor_contents for x in final_descriptors if isinstance(x, EDF.PluginDescriptor)])

        return self

    # This is a generator to reduce code duplication when wrapping the pathenv objects.
    def _get_paths(self):
        if self.paths is not None:
            # Apply in reverse order to get the expected hierarchy.
            for path_descriptor in reversed(self.paths):
                # Use the helper factory to get an object
                # capable of managing each dependency.
                yield EDF.PathEnv(path_descriptor)

    # This is a generator to reduce code duplication when wrapping the extdep objects.
    def _get_extdeps(self, env_object):
        if self.extdeps is not None:
            global_cache_path = env_object.get_shell_var("STUART_EXTDEP_CACHE_PATH")
            # Apply in reverse order to get the expected hierarchy.
            for extdep_descriptor in reversed(self.extdeps):
                # Use the helper factory to get an object
                # capable of managing each dependency.
                extdep = external_dependency.ExtDepFactory(extdep_descriptor)
                if global_cache_path is not None:
                    extdep.set_global_cache_path(global_cache_path)
                yield extdep

    def _apply_descriptor_object_to_env(self, desc_object, env_object):
        # Walk through each possible environment modification
        # and apply to the environment as required.

        if 'set_path' in desc_object.flags:
            env_object.insert_path(desc_object.published_path)
        if 'set_pypath' in desc_object.flags:
            env_object.insert_pypath(desc_object.published_path)
        if 'set_build_var' in desc_object.flags:
            env_object.set_build_var(
                desc_object.var_name, desc_object.published_path)
        if 'set_shell_var' in desc_object.flags:
            env_object.set_shell_var(
                desc_object.var_name, desc_object.published_path)

    def update_simple_paths(self, env_object):
        """Updates simple paths."""
        logging.debug("--- self_describing_environment.update_simple_paths()")
        for path in self._get_paths():
            self._apply_descriptor_object_to_env(path, env_object)

    def update_extdep_paths(self, env_object):
        """Updates external dependency paths."""
        logging.debug("--- self_describing_environment.update_extdep_paths()")
        for extdep in self._get_extdeps(env_object):
            self._apply_descriptor_object_to_env(extdep, env_object)

    def report_extdep_version(self, env_object):
        """Reports the version of all external dependencies."""
        logging.debug("--- self_describing_environment.report_extdep_version()")
        for extdep in self._get_extdeps(env_object):
            extdep.report_version()

    def update_extdeps(self, env_object):
        """Updates external dependencies."""
        logging.debug("--- self_describing_environment.update_extdeps()")
        # This function is called by our thread pool

        def update_extdep(self, extdep):
            # Check to see whether it's necessary to fetch the files.
            try:
                if not extdep.verify():
                    # Get rid of extdep published path since it could get changed
                    # during the fetch routine.
                    if 'set_path' in extdep.flags:
                        env_object.remove_path_element(extdep.published_path)
                    if 'set_pypath' in extdep.flags:
                        env_object.remove_pypath_element(extdep.published_path)
                    extdep.clean()
                    extdep.fetch()
                    # Re-apply the extdep to environment
                    self._apply_descriptor_object_to_env(extdep, env_object)
                return True
            except RuntimeError as e:
                logging.warning(f"[SDE] Failed to fetch {extdep}: {e}")
                if extdep.error_msg is not None:
                    logging.warning(extdep.error_msg)
                return False
            except FileNotFoundError:
                logging.warning(f"[SDE] Unable to fetch {extdep}")
                if extdep.error_msg is not None:
                    logging.warning(extdep.error_msg)
                return False
        # prep the worker pool
        all_extdeps = self._get_extdeps(env_object)
        self_extdeps = [(self, x) for x in all_extdeps]
        num_extdeps = len(self_extdeps)
        # if there are no ext_deps to update, bail early
        if num_extdeps == 0:
            return (0, 0)
        # don't create more threads than needed
        num_threads = min(os.cpu_count(), num_extdeps)
        # create a pool
        pool = dummy.Pool(num_threads)
        logging.debug(f"Creating {num_threads} threads for the SDE update")
        # map the task to the data
        pool_handle = pool.starmap_async(update_extdep, self_extdeps)
        pool.close()
        # use print so it doesn't go to the log
        print("Updating", end="", flush=True)
        old_count = num_extdeps
        # wait for the pool_handle (MapResult) to finish
        while pool_handle._number_left != 0:
            while (old_count != pool_handle._number_left and old_count > 0):
                print(".", end="", flush=True)
                old_count -= 1
            time.sleep(0.1)  # wait 100 ms
        print(". Done")
        # get the results
        results = pool_handle.get()
        success_count = results.count(True)
        failure_count = results.count(False)
        exception_count = results.count(None)
        if len(results) != num_extdeps or exception_count > 0:
            # We don't know where the error since we don't get a return result from it
            # so just tell users to check their logs
            raise RuntimeError("We encountered an exception while updating ext-deps. Review your log")
        return success_count, failure_count

    def clean_extdeps(self, env_object):
        """Cleans external dependencies."""
        for extdep in self._get_extdeps(env_object):
            extdep.clean()
            # TODO: Determine whether we want to update the env.

    def verify_extdeps(self, env_object):
        """Verifies external dependencies."""
        result = True
        for extdep in self._get_extdeps(env_object):
            if not extdep.verify():
                result = False
                logging.error("Dependency '%s' is not met!" % extdep.name)

        return result


def DestroyEnvironment():
    """Destroys global environment state."""
    global ENVIRONMENT_BOOTSTRAP_COMPLETE, ENV_STATE

    ENVIRONMENT_BOOTSTRAP_COMPLETE = False
    ENV_STATE = None


def BootstrapEnvironment(workspace, scopes=(), skipped_dirs=()):
    """Performs a multistage bootstrap of the environment.

    1. Locate and load all environment description files
    2. Parse all PATH-related descriptor files
    3. Load modules that had dependencies
    4. Report versions into the version aggregator

    Args:
        workspace (str): workspace root
        scopes (Tuple): scopes being built against
        skipped_dirs (Tuple): directories to ignore

    WARNING: if only one scope or skipped_dir, the tuple should end with a comma
    example: '(myscope,)'
    """
    global ENVIRONMENT_BOOTSTRAP_COMPLETE, ENV_STATE

    if not ENVIRONMENT_BOOTSTRAP_COMPLETE:
        #
        # ENVIRONMENT BOOTSTRAP STAGE 1
        # Locate and load all environment description files.
        #
        build_env = self_describing_environment(
            workspace, scopes, skipped_dirs).load_workspace()

        #
        # ENVIRONMENT BOOTSTRAP STAGE 2
        # Parse all of the PATH-related descriptor files to make sure that
        # any required tools or Python modules are now available.
        #
        shell_env = shell_environment.GetEnvironment()
        build_env.update_simple_paths(shell_env)

        #
        # ENVIRONMENT BOOTSTRAP STAGE 3
        # Now that the preliminary paths have been loaded,
        # we can load the modules that had greater dependencies.
        #
        build_env.update_extdep_paths(shell_env)

        #
        # ENVIRONMENT BOOTSTRAP STAGE 4
        # Report versions into the version aggregator
        build_env.report_extdep_version(shell_env)

        # Debug the environment that was produced.
        shell_env.log_environment()

        ENVIRONMENT_BOOTSTRAP_COMPLETE = True
        ENV_STATE = (build_env, shell_env)

    # Return the environment as it's configured.
    return ENV_STATE


def CleanEnvironment(workspace, scopes=(), skipped_dirs=()):
    """Cleans all external dependencies based on environment.

    Environment is bootstrapped from provided arguments and all dependencies
    are cleaned from that.

    Args:
        workspace (str): workspace root
        scopes (Tuple): scopes being built against
        skipped_dirs (Tuple): directories to ignore

    WARNING: if only one scope or skipped_dir, the tuple should end with a comma
    example: '(myscope,)'
    """
    # Bootstrap the environment.
    (build_env, shell_env) = BootstrapEnvironment(workspace, scopes, skipped_dirs)

    # Clean all the dependencies.
    build_env.clean_extdeps(shell_env)


def UpdateDependencies(workspace, scopes=(), skipped_dirs=()):
    """Updates all external dependencies based on environment.

    Environment is bootstrapped from provided arguments and all dependencies
    are updated from that.

    Args:
        workspace (str): workspace root
        scopes (Tuple): scopes being built against
        skipped_dirs (Tuple): directories to ignore

    WARNING: if only one scope or skipped_dir, the tuple should end with a comma
    example: '(myscope,)'
    """
    # Bootstrap the environment.
    (build_env, shell_env) = BootstrapEnvironment(workspace, scopes, skipped_dirs)

    # Update all the dependencies.
    return build_env.update_extdeps(shell_env)


def VerifyEnvironment(workspace, scopes=(), skipped_dirs=()):
    """Verifies all external dependencies based on environment.

    Environment is bootstrapped from provided arguments and all dependencies
    are verified from that.

    Args:
        workspace (str): workspace root
        scopes (Tuple): scopes being built against
        skipped_dirs (Tuple): directories to ignore

    WARNING: if only one scope or skipped_dir, the tuple should end with a comma
    example: '(myscope,)'
    """
    # Bootstrap the environment.
    (build_env, shell_env) = BootstrapEnvironment(workspace, scopes, skipped_dirs)

    # Verify all the dependencies.
    return build_env.verify_extdeps(shell_env)



# @file test_edk2_plat_build.py
# This contains unit tests for the edk2_plat_build
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
from edk2toolext.invocables.edk2_platform_build import Edk2PlatformBuild
import sys
import os
import logging
import shutil
from importlib import reload
from edk2toolext.environment import shell_environment
from edk2toolext.tests.uefi_tree import uefi_tree
from edk2toolext.environment import self_describing_environment
from edk2toolext.environment import version_aggregator


class TestEdk2PlatBuild(unittest.TestCase):

    minimalTree = None

    def setUp(self):
        TestEdk2PlatBuild.restart_logging()
        tree = uefi_tree()
        self.minimalTree = tree.get_workspace()
        pass

    def tearDown(self):
        shell_environment.GetEnvironment().restore_initial_checkpoint()
        buildFolder = os.path.join(self.minimalTree, "Build")
        shutil.rmtree(buildFolder, ignore_errors=True)
        TestEdk2PlatBuild.restart_logging()
        # we need to make sure to tear down the version aggregator and the SDE
        self_describing_environment.DestroyEnvironment()
        version_aggregator.ResetVersionAggregator()
        pass

    @classmethod
    def restart_logging(cls):
        '''
        We restart logging as logging is closed at the end of edk2 invocables.
        We also initialize it at the start.
        Reloading is the easiest way to get fresh state
        '''
        logging.shutdown()
        reload(logging)

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def test_init(self):
        builder = Edk2PlatformBuild()
        self.assertIsNotNone(builder)

    def test_ci_setup(self):
        builder = Edk2PlatformBuild()
        settings_file = os.path.join(self.minimalTree, "settings.py")
        sys.argv = ["stuart_build", "-c", settings_file]
        try:
            builder.Invoke()
        except SystemExit:
            # self.assertEqual(e.code, 0, "We should have a non zero error code")
            # we'll fail because we don't have the build command
            pass



# @file test_version_aggregator.py
# Unit test suite for the version_aggregator class.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
from edk2toolext.environment import version_aggregator


class TestVersionAggregator(unittest.TestCase):
    def setUp(self):
        version_aggregator.ResetVersionAggregator()
        pass

    def tearDown(self):
        pass

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def test_Singelton(self):
        version1 = version_aggregator.GetVersionAggregator()
        version2 = version_aggregator.GetVersionAggregator()
        self.assertEqual(version1, version2)
        self.assertIsNotNone(version1)
        version3 = version_aggregator.version_aggregator()
        version4 = version_aggregator.version_aggregator()
        self.assertNotEqual(version3, version4)

    def test_ReportVersions(self):
        version1 = version_aggregator.version_aggregator()
        version1.ReportVersion("test", "test", version_aggregator.VersionTypes.TOOL)
        self.assertEqual(len(version1.GetAggregatedVersionInformation()), 1)
        pass

    def test_ReportVersions_collision(self):
        version1 = version_aggregator.version_aggregator()
        version1.ReportVersion("test", "test", version_aggregator.VersionTypes.TOOL)
        # if keys don't collide we are good
        version1.ReportVersion("test2", "test", version_aggregator.VersionTypes.COMMIT)
        version1.ReportVersion("test3", "test", version_aggregator.VersionTypes.BINARY)
        version1.ReportVersion("test4", "test", version_aggregator.VersionTypes.INFO)
        # we're good to report the same thing twice as long as it matches
        version1.ReportVersion("test", "test", version_aggregator.VersionTypes.TOOL)
        with self.assertRaises(ValueError):
            version1.ReportVersion("test", "test2", version_aggregator.VersionTypes.INFO)

    def test_GetInformation(self):
        version1 = version_aggregator.version_aggregator()
        test_ver = ["I think", "I exist"]
        version1.ReportVersion("I think", "therefore", version_aggregator.VersionTypes.INFO)
        version1.ReportVersion("I exist", "proof", version_aggregator.VersionTypes.INFO)
        test2_ver = list(version1.GetAggregatedVersionInformation().keys())
        self.assertListEqual(test_ver, test2_ver)
        version1.ReportVersion("test", "test", version_aggregator.VersionTypes.TOOL)
        self.assertEqual(len(test2_ver), 2)
        test2_ver = list(version1.GetAggregatedVersionInformation().keys())
        self.assertEqual(len(test2_ver), 3)

    def test_reset(self):
        version1 = version_aggregator.version_aggregator()
        version1.ReportVersion("test", "I exist", version_aggregator.VersionTypes.INFO)
        self.assertEqual(len(version1.GetAggregatedVersionInformation()), 1)
        version1.Reset()
        self.assertEqual(len(version1.GetAggregatedVersionInformation()), 0)

    def test_global_reset(self):
        version1 = version_aggregator.version_aggregator()
        version1.ReportVersion("test", "I exist", version_aggregator.VersionTypes.INFO)
        self.assertEqual(len(version1.GetAggregatedVersionInformation()), 1)
        version1 = version_aggregator.version_aggregator()
        version_aggregator.ResetVersionAggregator()
        self.assertEqual(len(version1.GetAggregatedVersionInformation()), 0)


if __name__ == '__main__':
    unittest.main()



# @file Edk2CiBuild.py
# This module contains code that supports CI/CD
# This is the main entry for the build and test process
# of Non-Product builds
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code that supports CI/CD via the ci_build invocable.

Contains a CIBuildSettingsManager that must be subclassed in a build settings
file. This provides platform specific information to Edk2CiBuild invocable
while allowing the invocable itself to remain platform agnostic.
"""
import os
import sys
import logging
import yaml
import traceback
from typing import Dict, Any
from edk2toollib.uefi.edk2.path_utilities import Edk2Path
from edk2toollib.log.junit_report_format import JunitTestReport
from edk2toolext.invocables.edk2_multipkg_aware_invocable import Edk2MultiPkgAwareInvocable
from edk2toolext.invocables.edk2_multipkg_aware_invocable import MultiPkgAwareSettingsInterface
from edk2toolext.environment import self_describing_environment
from edk2toolext.environment.plugintypes.ci_build_plugin import ICiBuildPlugin
from edk2toolext.environment import shell_environment
from edk2toolext import edk2_logging


class CiBuildSettingsManager(MultiPkgAwareSettingsInterface):
    """Platform specific settings for Edk2CiBuild.

    Provide information necessary for `stuart_ci_build.exe` or
    `edk2_ci_build.py` to successfully execute.

    Example: Example: Overriding CiBuildSettingsManager
        ```python
        from edk2toolext.invocables.edk2_ci_build import CiBuildSettingsManager
        import yaml
        class CiManager(CiBuildSettingsManager):
            def GetDependencies(self):
                return {
                    "Path": "/Common/MU",
                    "Url":  "https://github.com/Microsoft/mu_tiano_plus.git"
                }
        ```
    """

    def GetName(self) -> str:
        """Get the name of the repo, platform, or product being build by CI.

        TIP: Required Override in a subclass

        Returns:
            (str): repo, platform, product
        """
        raise NotImplementedError()

    def GetPluginSettings(self) -> Dict[str, Any]:
        """Provide a dictionary of global settings for individual plugins.

        TIP: Optional Override in a subclass

        WARNING:
            This sets the global plugin configurations. Edk2CiBuild automatically searches for,
            and loads, the package ci settings file if it exists. This file will override these
            settings. This file must be located at the base of the package named <Package>.ci.yaml.

            Ex: EmbeddedPkg/EmbeddedPkg.ci.yaml.

        Returns:
            (Dict[str, Any]): plugin settings
        """
        return {}


class Edk2CiBuild(Edk2MultiPkgAwareInvocable):
    """Invocable supporting an iterative multi-package build and test process leveraging CI build plugins."""

    def GetSettingsClass(self):
        """Returns the CiBuildSettingsManager class.

        WARNING: CiBuildSettingsManager must be subclassed in your platform settings file.
        """
        return CiBuildSettingsManager

    def GetLoggingFileName(self, loggerType):
        """Returns the filename (CI_BUILDLOG) of where the logs for the Edk2CiBuild invocable are stored in."""
        return "CI_BUILDLOG"

    def Go(self):
        """Executes the core functionality of the Edk2CiBuild invocable."""
        log_directory = os.path.join(self.GetWorkspaceRoot(), self.GetLoggingFolderRelativeToRoot())

        Edk2CiBuild.collect_python_pip_info()

        # make Edk2Path object to handle all path operations
        try:
            edk2path = Edk2Path(self.GetWorkspaceRoot(), self.PlatformSettings.GetPackagesPath())
        except Exception as e:
            logging.error("Src Tree is invalid.  Did you Setup correctly?")
            raise e

        logging.info(f"Running CI Build: {self.PlatformSettings.GetName()}")
        logging.info(f"WorkSpace: {edk2path.WorkspacePath}")
        logging.info(f"Package Path: {os.pathsep.join(edk2path.PackagePathList)}")
        # Bring up the common minimum environment.
        logging.log(edk2_logging.SECTION, "Getting Environment")
        (build_env, shell_env) = self_describing_environment.BootstrapEnvironment(
            self.GetWorkspaceRoot(), self.GetActiveScopes(), self.GetSkippedDirectories())
        env = shell_environment.GetBuildVars()

        # Bind our current execution environment into the shell vars.
        ph = os.path.dirname(sys.executable)
        if " " in ph:
            ph = '"' + ph + '"'
        shell_env.set_shell_var("PYTHON_HOME", ph)
        # PYTHON_COMMAND is required to be set for using edk2 python builds.
        # todo: work with edk2 to remove the bat file and move to native python calls.
        #       This would be better in an edk2 plugin so that it could be modified/controlled
        #       more easily
        #
        pc = sys.executable
        if " " in pc:
            pc = '"' + pc + '"'
        shell_env.set_shell_var("PYTHON_COMMAND", pc)

        env.SetValue("TARGET_ARCH", " ".join(self.requested_architecture_list), "from edk2 ci build.py")

        # Generate consumable XML object- junit format
        JunitReport = JunitTestReport()

        # Keep track of failures
        failure_num = 0
        total_num = 0

        # Load plugins
        logging.log(edk2_logging.SECTION, "Loading plugins")

        pluginList = self.plugin_manager.GetPluginsOfClass(ICiBuildPlugin)

        for pkgToRunOn in self.requested_package_list:
            #
            # run all loaded Edk2CiBuild Plugins/Tests
            #
            logging.log(edk2_logging.SECTION, f"Building {pkgToRunOn} Package")
            logging.info(f"Running on Package: {pkgToRunOn}")
            package_class_name = f"Edk2CiBuild.{self.PlatformSettings.GetName()}.{pkgToRunOn}"
            ts = JunitReport.create_new_testsuite(pkgToRunOn, package_class_name)
            packagebuildlog_path = os.path.join(log_directory, pkgToRunOn)
            _, txt_handle = edk2_logging.setup_txt_logger(
                packagebuildlog_path, f"BUILDLOG_{pkgToRunOn}", logging_level=logging.DEBUG, isVerbose=True)
            _, md_handle = edk2_logging.setup_markdown_logger(
                packagebuildlog_path, f"BUILDLOG_{pkgToRunOn}", logging_level=logging.DEBUG, isVerbose=True)
            loghandle = [txt_handle, md_handle]
            shell_environment.CheckpointBuildVars()
            env = shell_environment.GetBuildVars()

            # load the package level .ci.yaml
            pkg_config_file = edk2path.GetAbsolutePathOnThisSystemFromEdk2RelativePath(
                os.path.join(pkgToRunOn, pkgToRunOn + ".ci.yaml"))
            if (pkg_config_file):
                with open(pkg_config_file, 'r') as f:
                    pkg_config = yaml.safe_load(f)
            else:
                logging.info(f"No Pkg Config file for {pkgToRunOn}")
                pkg_config = dict()

            # get all the defines from the package configuration
            if "Defines" in pkg_config:
                for definition_key in pkg_config["Defines"]:
                    definition = pkg_config["Defines"][definition_key]
                    env.SetValue(definition_key, definition, "Edk2CiBuild.py from PkgConfig yaml", False)

            # For each plugin
            for Descriptor in pluginList:
                # For each target
                for target in self.requested_target_list:

                    if (target not in Descriptor.Obj.RunsOnTargetList()):
                        continue

                    edk2_logging.log_progress(f"--Running {pkgToRunOn}: {Descriptor.Name} {target} --")
                    total_num += 1
                    shell_environment.CheckpointBuildVars()
                    env = shell_environment.GetBuildVars()

                    env.SetValue("TARGET", target, "Edk2CiBuild.py before RunBuildPlugin")
                    (testcasename, testclassname) = Descriptor.Obj.GetTestName(package_class_name, env)
                    tc = ts.create_new_testcase(testcasename, testclassname)

                    # create the stream for the build log
                    plugin_output_stream = edk2_logging.create_output_stream()

                    # merge the repo level and package level for this specific plugin
                    pkg_plugin_configuration = Edk2CiBuild.merge_config(self.PlatformSettings.GetPluginSettings(),
                                                                        pkg_config, Descriptor.descriptor)

                    # Still need to see if the package decided this should be skipped
                    if pkg_plugin_configuration is None or\
                            "skip" in pkg_plugin_configuration and pkg_plugin_configuration["skip"]:
                        tc.SetSkipped()
                        edk2_logging.log_progress("--->Test Skipped by package! %s" % Descriptor.Name)

                    else:
                        try:
                            #   - package is the edk2 path to package.  This means workspace/package path relative.
                            #   - edk2path object configured with workspace and packages path
                            #   - any additional command line args
                            #   - RepoConfig Object (dict) for the build
                            #   - PkgConfig Object (dict)
                            #   - EnvConfig Object
                            #   - Plugin Manager Instance
                            #   - Plugin Helper Obj Instance
                            #   - testcase Object used for outputing junit results
                            #   - output_stream the StringIO output stream from this plugin
                            rc = Descriptor.Obj.RunBuildPlugin(pkgToRunOn, edk2path, pkg_plugin_configuration,
                                                               env, self.plugin_manager, self.helper,
                                                               tc, plugin_output_stream)
                        except Exception as exp:
                            exc_type, exc_value, exc_traceback = sys.exc_info()
                            logging.critical("EXCEPTION: {0}".format(exp))
                            exceptionPrint = traceback.format_exception(type(exp), exp, exc_traceback)
                            logging.critical(" ".join(exceptionPrint))
                            tc.SetError("Exception: {0}".format(
                                exp), "UNEXPECTED EXCEPTION")
                            rc = 1

                        if (rc > 0):
                            failure_num += 1
                            if (rc is None):
                                logging.error(
                                    f"--->Test Failed: {Descriptor.Name} {target} returned NoneType")
                            else:
                                logging.error(
                                    f"--->Test Failed: {Descriptor.Name} {target} returned {rc}")
                        elif (rc < 0):
                            logging.warn(f"--->Test Skipped: in plugin! {Descriptor.Name} {target}")
                        else:
                            edk2_logging.log_progress(f"--->Test Success: {Descriptor.Name} {target}")

                    # revert to the checkpoint we created previously
                    shell_environment.RevertBuildVars()
                    # remove the logger
                    edk2_logging.remove_output_stream(plugin_output_stream)
                # finished target loop
            # Finished plugin loop

            edk2_logging.stop_logging(loghandle)  # stop the logging for this particular buildfile
            shell_environment.RevertBuildVars()
        # Finished buildable file loop

        JunitReport.Output(os.path.join(self.GetWorkspaceRoot(), "Build", "TestSuites.xml"))

        # Print Overall Success
        if (failure_num != 0):
            logging.error("Overall Build Status: Error")
            edk2_logging.log_progress(f"There were {failure_num} failures out of {total_num} attempts")
        else:
            edk2_logging.log_progress("Overall Build Status: Success")

        return failure_num

    @staticmethod
    def merge_config(gbl_config, pkg_config, descriptor={}):
        """Merge two configurations.

        One global and one specificto the package to create the proper config
        for a plugin to execute.

        Returns:
            (dict): Dictionary of config settings
        """
        plugin_name = ""
        config = dict()
        if "module" in descriptor:
            plugin_name = descriptor["module"]
        if "config_name" in descriptor:
            plugin_name = descriptor["config_name"]

        if plugin_name == "":
            return config

        if plugin_name in gbl_config:
            config.update(gbl_config[plugin_name])

        if plugin_name in pkg_config:
            config.update(pkg_config[plugin_name])

        return config


def main():
    """Entry point to invoke Edk2CiBuild."""
    Edk2CiBuild().Invoke()



## @file test_git_dependency.py
# Unit test suite for the GitDependency class.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import os
import unittest
import logging
import shutil
import stat
import tempfile
import copy
from edk2toolext.environment import environment_descriptor_files as EDF
from edk2toolext.environment.extdeptypes.git_dependency import GitDependency
from edk2toolext.environment import shell_environment

test_dir = None
uptodate_version = "7fd1a60b01f91b314f59955a4e4d4e80d8edf11d"
behind_one_version = "762941318ee16e59dabbacb1b4049eec22f0d303"
invalid_version = "762941318ee16e59d123456789049eec22f0d303"

hw_json_template = '''
{
  "scope": "global",
  "type": "git",
  "name": "HelloWorld",
  "source": "https://github.com/octocat/Hello-World.git",
  "version": "%s",
  "flags": []
}
'''


def prep_workspace():
    global test_dir
    # if test temp dir doesn't exist
    if test_dir is None or not os.path.isdir(test_dir):
        test_dir = tempfile.mkdtemp()
        logging.debug("temp dir is: %s" % test_dir)
    else:
        clean_workspace()
        test_dir = tempfile.mkdtemp()


def clean_workspace():
    global test_dir
    if test_dir is None:
        return

    if os.path.isdir(test_dir):

        # spell-checker:ignore dorw
        def dorw(action, name, exc):
            os.chmod(name, stat.S_IWRITE)
            if (os.path.isdir(name)):
                os.rmdir(name)
            else:
                os.remove(name)

        shutil.rmtree(test_dir, onerror=dorw)
        test_dir = None


class TestGitDependency(unittest.TestCase):
    def setUp(self):
        prep_workspace()

    @classmethod
    def setUpClass(cls):
        logger = logging.getLogger('')
        logger.addHandler(logging.NullHandler())
        unittest.installHandler()

    @classmethod
    def tearDownClass(cls):
        clean_workspace()

    # good case
    def test_fetch_verify_good_repo_at_top_of_tree(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % uptodate_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        ext_dep.fetch()
        self.assertTrue(ext_dep.verify())
        self.assertEqual(ext_dep.version, uptodate_version)

    def test_fetch_verify_good_repo_at_not_top_of_tree(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % behind_one_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        ext_dep.fetch()
        self.assertTrue(ext_dep.verify())
        self.assertEqual(ext_dep.version, behind_one_version)

    def test_fetch_verify_non_existant_repo_commit_hash(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % invalid_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        ext_dep.fetch()
        self.assertEqual(ext_dep.version, invalid_version)
        self.assertFalse(ext_dep.verify(), "Should not verify")

    def test_verify_no_directory(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % invalid_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        self.assertFalse(ext_dep.verify())

    def test_verify_empty_repo_dir(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % invalid_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        os.makedirs(ext_dep._local_repo_root_path, exist_ok=True)
        self.assertFalse(ext_dep.verify())

    def test_verify_invalid_git_repo(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % invalid_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        os.makedirs(ext_dep._local_repo_root_path, exist_ok=True)
        with open(os.path.join(ext_dep._local_repo_root_path, "testfile.txt"), 'a') as my_file:
            my_file.write("Test code\n")
        self.assertFalse(ext_dep.verify())

    def test_verify_dirty_git_repo(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % uptodate_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        ext_dep.fetch()
        # now write a new file
        with open(os.path.join(ext_dep._local_repo_root_path, "testfile.txt"), 'a') as my_file:
            my_file.write("Test code to make repo dirty\n")
        self.assertFalse(ext_dep.verify())

    def test_verify_up_to_date(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % uptodate_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        ext_dep.fetch()
        self.assertTrue(ext_dep.verify())

    def test_verify_down_level_repo(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % behind_one_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        ext_dep.fetch()
        self.assertTrue(ext_dep.verify(), "Confirm valid ext_dep at one commit behind")

        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % uptodate_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        self.assertFalse(ext_dep.verify(), "Confirm downlevel repo fails to verify")
        ext_dep.fetch()
        self.assertTrue(ext_dep.verify(), "Confirm repo can be updated")

    # CLEAN TESTS

    def test_clean_no_directory(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % uptodate_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        self.assertFalse(os.path.isdir(ext_dep.contents_dir), "Confirm not ext dep directory before cleaning")
        ext_dep.clean()
        self.assertFalse(os.path.isdir(ext_dep.contents_dir))

    def test_clean_dir_but_not_git_repo(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % invalid_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        os.makedirs(ext_dep._local_repo_root_path, exist_ok=True)
        with open(os.path.join(ext_dep._local_repo_root_path, "testfile.txt"), 'a') as my_file:
            my_file.write("Test code\n")
        ext_dep.clean()
        self.assertFalse(os.path.isdir(ext_dep.contents_dir))

    def test_clean_dirty_git_repo(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % uptodate_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        ext_dep.fetch()
        self.assertTrue(ext_dep.verify(), "Confirm repo is valid")
        # now write a new file
        with open(os.path.join(ext_dep._local_repo_root_path, "testfile.txt"), 'a') as my_file:
            my_file.write("Test code to make repo dirty\n")
        self.assertFalse(ext_dep.verify(), "Confirm repo is dirty")
        ext_dep.clean()
        self.assertFalse(os.path.isdir(ext_dep.contents_dir))

    def test_clean_clean_repo(self):
        ext_dep_file_path = os.path.join(test_dir, "hw_ext_dep.json")
        with open(ext_dep_file_path, "w+") as ext_dep_file:
            ext_dep_file.write(hw_json_template % uptodate_version)

        ext_dep_descriptor = EDF.ExternDepDescriptor(ext_dep_file_path).descriptor_contents
        ext_dep = GitDependency(ext_dep_descriptor)
        ext_dep.fetch()
        self.assertTrue(ext_dep.verify(), "Confirm repo is valid and clean")
        ext_dep.clean()
        self.assertFalse(os.path.isdir(ext_dep.contents_dir))


class TestGitDependencyUrlPatching(unittest.TestCase):
    TEST_DESCRIPTOR = {
        "descriptor_file": os.path.abspath(__file__),
        "scope": "global",
        "type": "git",
        "name": "HelloWorld",
        "source": "https://github.com/octocat/Hello-World.git",
        "version": "7fd1a60b01f91b314f59955a4e4d4e80d8edf11d",
        "flags": []
    }

    def tearDown(self):
        env = shell_environment.GetEnvironment()
        env.restore_checkpoint(TestGitDependencyUrlPatching.env_checkpoint)

    @classmethod
    def setUpClass(cls):
        env = shell_environment.GetEnvironment()
        cls.env_checkpoint = env.checkpoint()

    #
    # URL FORMATTING TESTS
    #
    def test_url_should_not_be_modified_without_env(self):
        my_test_descriptor = copy.copy(TestGitDependencyUrlPatching.TEST_DESCRIPTOR)
        # Add the indicator for patching.
        my_test_descriptor['url_creds_var'] = 'test_creds_var'

        # Initialize the GitDependency object.
        git_dep = GitDependency(my_test_descriptor)

        # Assert that the URL is identical.
        self.assertEqual(git_dep.source, my_test_descriptor['source'])

    def test_url_should_not_be_modified_without_descriptor_field(self):
        my_test_descriptor = copy.copy(TestGitDependencyUrlPatching.TEST_DESCRIPTOR)

        env = shell_environment.GetEnvironment()
        # Add the var to the environment.
        env.set_shell_var('test_creds_var', 'my_stuff')

        # Initialize the GitDependency object.
        git_dep = GitDependency(my_test_descriptor)

        # Assert that the URL is identical.
        self.assertEqual(git_dep.source, my_test_descriptor['source'])

    def test_url_should_be_modified_if_creds_are_indicated_and_supplied(self):
        my_test_descriptor = copy.copy(TestGitDependencyUrlPatching.TEST_DESCRIPTOR)
        # Add the indicator for patching.
        my_test_descriptor['url_creds_var'] = 'test_creds_var'

        env = shell_environment.GetEnvironment()
        # Add the var to the environment.
        env.set_shell_var('test_creds_var', 'my_stuff')

        # Initialize the GitDependency object.
        git_dep = GitDependency(my_test_descriptor)

        # Assert that the URL is identical.
        self.assertEqual(git_dep.source, "https://my_stuff@github.com/octocat/Hello-World.git")


if __name__ == '__main__':
    unittest.main()



# @file test_versioninfo.py
# This unittest module contains test cases for the versioninfo_tool module and CLI routines.
#
##
# Copyright (C) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import os
import unittest
import tempfile
import json
import copy
import logging
import shutil
from io import StringIO
from edk2toolext.versioninfo import versioninfo_tool
from edk2toollib.utility_functions import RunCmd

DUMMY_EXE_FILE_NAME = 'dummy_exe'
DUMMY_JSON_FILE_NAME = 'dummy_json_file'
VERSIONINFO_JSON_FILE_NAME = 'VERSIONINFO'
BAD_JSON_FILE_NAME = 'bad_json'

DUMMY_VALID_JSON = {
    "Minimal": "False",
    "FileVersion": "1.2.3.4",
    "ProductVersion": "1.2.3.4",
    "FileFlagsMask": "0x3f",
    "FileFlags": "0x0",
    "FileOS": "VOS_NT",
    "FileType": "VFT_DRV",
    "FileSubtype": "VFT2_DRV_SYSTEM",
    "StringFileInfo": {
        "Comments": "Dummy Driver",
        "CompanyName": "Dummy company",
        "FileDescription": "Dummy Driver",
        "FileVersion": "1.2.3.4",
        "InternalName": "Test Name",
        "LegalCopyright": "(C) 2048 Dummy Driver",
        "OriginalFilename": "Dummy.sys",
        "ProductName": "Dummy Driver"
    },
    "VarFileInfo": {
        "Translation": "0x0409 0x04b0"
    }
}

DUMMY_MINIMAL_JSON = {
    "Fileversion": "1,0,0,0",
    "OriginalFilename": "Test Name",
    "CompanyName": "Test Company"
}

DUMMY_MINIMAL_DECODED = {
    "FileVersion": "1.0.0.0",
    "ProductVersion": "0.0.0.0",
    "FileFlagsMask": "0x0",
    "FileFlags": "0x0",
    "FileOS": "VOS_UNKNOWN",
    "FileType": "VFT_UNKNOWN",
    "FileSubtype": "VFT2_UNKNOWN",
    "StringFileInfo": {
        "CompanyName": "Test Company",
        "OriginalFilename": "Test Name"
    },
    "VarFileInfo": {
        "Translation": "0x0409 0x04b0"
    }
}


def check_for_err_helper(cls, temp_dir, json_input, err_msg, decode=False):
    with StringIO() as log_stream:
        log_handler = logging.StreamHandler(log_stream)
        log_handler.setLevel(logging.DEBUG)  # set the logger to get everything
        logging.getLogger().addHandler(log_handler)
        returned_error = False

        if decode:
            returned_error = not versioninfo_tool.decode_version_info_dump_json(json_input, temp_dir)
        else:
            returned_error = not versioninfo_tool.encode_version_info_dump_rc(json_input, temp_dir)

        logging.getLogger().removeHandler(log_handler)
        cls.assertFalse(os.path.isfile(os.path.join(temp_dir, 'VERSIONINFO.rc')))
        cls.assertTrue(err_msg in log_stream.getvalue())
        cls.assertTrue(returned_error)


def compared_decoded_version_info(self, json_file_path, reference):
    try:
        generated_json = open(json_file_path)
        generated_dict = json.load(generated_json)
        self.assertTrue('Signature' in generated_dict)
        del generated_dict['Signature']
        self.assertTrue('StrucVersion' in generated_dict)
        del generated_dict['StrucVersion']
        if 'FileDateMS' in generated_dict:
            del generated_dict['FileDateMS']
        if 'FileDateLS' in generated_dict:
            del generated_dict['FileDateLS']
        ref = copy.deepcopy(reference)
        if "Minimal" in ref:
            del ref["Minimal"]
        self.assertEqual(generated_dict, ref)
    except ValueError:
        self.fail()
    except IOError:
        self.fail()


class TestVersioninfo(unittest.TestCase):

    def test_encode_decode_full(self):
        temp_dir = tempfile.mkdtemp()
        # Create the EXE file
        versioned_exe_path = os.path.join(temp_dir, DUMMY_EXE_FILE_NAME) + '.exe'
        source_exe_path = os.path.join(os.path.dirname(__file__), "testdata", "versioninfo_full_exe.data")
        shutil.copyfile(source_exe_path, versioned_exe_path)
        # Create the parameters that will go to the service request function
        version_info_output_path = os.path.join(temp_dir, VERSIONINFO_JSON_FILE_NAME + '.json')
        versioninfo_tool.decode_version_info_dump_json(versioned_exe_path, version_info_output_path)

        # then we compare to make sure it matches what it should be
        compared_decoded_version_info(self, version_info_output_path, DUMMY_VALID_JSON)

    def test_encode_decode_minimal(self):
        temp_dir = tempfile.mkdtemp()
        # Create the EXE file
        versioned_exe_path = os.path.join(temp_dir, DUMMY_EXE_FILE_NAME) + '.exe'
        source_exe_path = os.path.join(os.path.dirname(__file__), "testdata", "versioninfo_minimal_exe.data")
        shutil.copyfile(source_exe_path, versioned_exe_path)
        # Create the parameters that will go to the service request function
        version_info_output_path = os.path.join(temp_dir, VERSIONINFO_JSON_FILE_NAME + '.json')
        versioninfo_tool.decode_version_info_dump_json(versioned_exe_path, version_info_output_path)

        # then we compare to make sure it matches what it should be
        compared_decoded_version_info(self, version_info_output_path, DUMMY_MINIMAL_DECODED)

    def test_encode_minimal(self):
        temp_dir = tempfile.mkdtemp()
        # Create the parameters that will go to the service request function
        version_info_input_path = os.path.join(temp_dir, "in.json")
        with open(version_info_input_path, "w") as f:
            json.dump(DUMMY_MINIMAL_JSON, f)
        obj = versioninfo_tool.encode_version_info(version_info_input_path)
        self.assertTrue(obj.validate_minimal())

    def test_missing_varinfo(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        del bad_json['VarFileInfo']
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Missing required parameter: VARFILEINFO')

    def test_missing_translation(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        del bad_json['VarFileInfo']['Translation']
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Missing required parameter in VarFileInfo: Translation')

    def test_invalid_varfileinfo(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['VarFileInfo']['FileVersion'] = "1.2.1.2"
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Invalid VarFileInfo parameter: FileVersion')

    def test_missing_companyname(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        del bad_json['StringFileInfo']['CompanyName']
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Missing required StringFileInfo parameter: CompanyName')

    def test_version_overflow(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['ProductVersion'] = '65536.0.0.0'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Integer overflow in version string')

    def test_invalid_version(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['ProductVersion'] = 'Version 1.0.1.0'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Version must be in form " INTEGER.INTEGER.INTEGER.INTEGER"')  # noqa

    def test_bad_version_format(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['ProductVersion'] = '1.234'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Version must be in form "INTEGER.INTEGER.INTEGER.INTEGER"')  # noqa

    def test_invalid_language_code_value(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['VarFileInfo']['Translation'] = '"0x0009 0x04b0"'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file,
                             f"Invalid language code: {bad_json['VarFileInfo']['Translation']}")

    def test_invalid_language_code_string(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['VarFileInfo']['Translation'] = '"utf-8 US"'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file,
                             f"Invalid language code: {bad_json['VarFileInfo']['Translation']}")

    def test_invalid_language_code_format(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['VarFileInfo']['Translation'] = '"0x400904b0"'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Translation field must contain 2 space delimited hexidecimal bytes')  # noqa

    def test_invalid_language_code_no_hex(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['VarFileInfo']['Translation'] = '"4009 04b0"'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Invalid language code: '
                             + bad_json['VarFileInfo']['Translation'])

    def test_invalid_fileos_hex(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['FileOS'] = '0x12391'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Invalid FILEOS value: ' + bad_json['FileOS'])

    def test_invalid_fileos_string(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['FileOS'] = 'INVALID'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Invalid FILEOS value: ' + bad_json['FileOS'])

    def test_invalid_filetype_hex(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['FileType'] = '0x12391'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Invalid FILESUBTYPE value for FILETYPE ' + bad_json['FileType'])  # noqa

    def test_invalid_filetype_string(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['FileType'] = 'INVALID'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Invalid FILESUBTYPE value for FILETYPE ' + bad_json['FileType'])  # noqa

    def test_invalid_filesubtype_drv(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['FileType'] = 'VFT_DRV'
        bad_json['FileSubtype'] = 'VFT2_FONT_RASTER'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Invalid FILESUBTYPE value for FILETYPE VFT_DRV: VFT2_FONT_RASTER')  # noqa

    def test_invalid_filesubtype_font(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        bad_json['FileType'] = 'VFT_FONT'
        bad_json['FileSubtype'] = 'VFT2_DRV_SOUND'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Invalid FILESUBTYPE value for FILETYPE VFT_FONT: VFT2_DRV_SOUND')  # noqa

    def test_missing_filetype(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        del bad_json['FileType']
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file,
                             'Missing parameter: must have FileType if FileSubtype defined')

    def test_no_stringinfo_header(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = copy.deepcopy(DUMMY_VALID_JSON)
        err_str = ''
        for key in bad_json['StringFileInfo']:
            if key == 'FileVersion':
                continue

            bad_json[key] = bad_json['StringFileInfo'][key]
            err_str += 'Invalid parameter: ' + key.upper() + '.\n'

        del bad_json['StringFileInfo']
        err_str += 'Missing required parameter: STRINGFILEINFO'
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, err_str)

    def test_bad_input_path(self):
        ''' check to make sure we throw an exception '''
        try:
            check_for_err_helper(self, ".", "bad/path/to/file.json", "Could not find bad/path/to/file.json\n")
            self.fail("We shouldn't have found the file")
        except FileNotFoundError:
            pass

    def test_command_line_interface(self):
        ''' makes sure the command line interface is working correctly '''
        ret = RunCmd('versioninfo_tool', '-h', logging_level=logging.ERROR)
        self.assertEqual(ret, 0)

    def test_non_pe_file(self):
        temp_dir = tempfile.mkdtemp()
        bad_pe = os.path.join(temp_dir, DUMMY_JSON_FILE_NAME + '.bad')

        with open(bad_pe, 'w') as bad_file:
            json.dump(DUMMY_VALID_JSON, bad_file)

        check_for_err_helper(self, temp_dir, bad_pe, "DOS Header magic not found", True)

    def test_invalid_json_format1(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = """
            "FileVersion": "1.2.3.4",
            "ProductVersion": "1.2.3.4",
            "FileFlagsMask": "0x3f",
            "FileFlags": "0x0",
            "FileOS": "VOS_NT",
            "FileType": "VFT_DRV",
            "FileSubtype": "VFT2_DRV_SYSTEM",
            "StringFileInfo": {
                "Comments": "Dummy Driver",
                "CompanyName": "Dummy compnay",
                "FileDescription": "Dummy Driver",
                "FileVersion": "1.2.3.4",
                "InternalName": "Internal Dummy Driver",
                "LegalCopyright": "(C) 2048 Dummy Driver",
                "OriginalFilename": "Dummy.sys",
                "ProductName": "Dummy Driver"
            },
            "VarFileInfo": {
                "Translation": "0x0409 0x04b0"
            }
        """

        with open(bad_json_file, 'w') as bad_file:
            bad_file.write(bad_json)

        check_for_err_helper(self, temp_dir, bad_json_file, "Invalid JSON format, Extra data")

    def test_invalid_json_format2(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = """ {
            FileVersion: 1.2.3.4,
            ProductVersion: 1.2.3.4,
            FileFlagsMask: 0x3f,
            FileFlags: 0x0,
            FileOS: VOS_NT,
            FileType: VFT_DRV,
            "FileSubtype": "VFT2_DRV_SYSTEM",
            "StringFileInfo": {
                "Comments": "Dummy Driver",
                "CompanyName": "Dummy compnay",
                "FileDescription": "Dummy Driver",
                "FileVersion": "1.2.3.4",
                "InternalName": "Internal Dummy Driver",
                "LegalCopyright": "(C) 2048 Dummy Driver",
                "OriginalFilename": "Dummy.sys",
                "ProductName": "Dummy Driver"
            },
            "VarFileInfo": {
                "Translation": "0x0409 0x04b0"
            }
        }
        """

        with open(bad_json_file, 'w') as bad_file:
            bad_file.write(bad_json)

        check_for_err_helper(self, temp_dir, bad_json_file, "Invalid JSON format, Expecting property name enclosed in double quotes")  # noqa

    def test_invalid_json_format3(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = """ {
            "FileVersion": "1.2.3.4",
            "ProductVersion": "1.2.3.4",
            "FileFlagsMask": "0x3f",
            "FileFlags": "0x0",
            "FileOS": "VOS_NT",
            "FileType": "VFT_DRV",
            "FileSubtype": "VFT2_DRV_SYSTEM",
            "StringFileInfo": [
                "Comments": "Dummy Driver",
                "CompanyName": "Dummy compnay",
                "FileDescription": "Dummy Driver",
                "FileVersion": "1.2.3.4",
                "InternalName": "Internal Dummy Driver",
                "LegalCopyright": "(C) 2048 Dummy Driver",
                "OriginalFilename": "Dummy.sys",
                "ProductName": "Dummy Driver"
            ],
            "VarFileInfo": [
                "Translation": "0x0409 0x04b0"
            ]
        }
        """

        with open(bad_json_file, 'w') as bad_file:
            bad_file.write(bad_json)

        check_for_err_helper(self, temp_dir, bad_json_file, "Invalid JSON format, Expecting ',' delimiter")  # noqa

    def test_invalid_minimal_fields(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = {
            "Minimal": "True",
            "FileVersion": "1.2.3.4",
            "CompanyName": "Test Company",
            "FileType": "VFT_DRV",
        }
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, 'Invalid minimal parameter: FILETYPE')

    def test_invalid_minimal_value(self):
        temp_dir = tempfile.mkdtemp()
        bad_json_file = os.path.join(temp_dir, BAD_JSON_FILE_NAME + '.json')
        bad_json = {
            "Minimal": "Yes",
            "FileVersion": "1.2.3.4",
            "CompanyName": "Test Company",
            "FileType": "VFT_DRV",
        }
        with open(bad_json_file, 'w') as bad_file:
            json.dump(bad_json, bad_file)

        check_for_err_helper(self, temp_dir, bad_json_file, "Invalid value for 'Minimal', must be boolean.")



## @file capsule_helper.py
# This module contains helper functions for building EDK2 FMP UEFI capsules from
# binary payloads, along with the functions to standardize the creation of the Windows
# driver installation files.
#
##
# Copyright (C) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Helper Functions for building EDK2 FMP UEFI Capsules.

This module contains helper functions for building EDK2 FMP UEFI capsules from
binary payloads, along with the functions to standardize the creation of the Windows
driver installation files
"""
import uuid
import os
import struct
import datetime

from typing import List
from dataclasses import dataclass
from dataclasses import field

from edk2toollib.windows.capsule import inf_generator2, cat_generator
from edk2toollib.uefi.uefi_capsule_header import UefiCapsuleHeaderClass
from edk2toollib.uefi.fmp_capsule_header import FmpCapsuleHeaderClass, FmpCapsuleImageHeaderClass
from edk2toollib.uefi.fmp_auth_header import FmpAuthHeaderClass
from edk2toollib.uefi.edk2.fmp_payload_header import FmpPayloadHeaderClass

# https://docs.microsoft.com/en-us/dotnet/api/system.security.cryptography.pkcs.contentinfo.-ctor?view=netframework-4.8
PKCS7_SIGNED_DATA_OID = '1.2.840.113549.1.7.2'


@dataclass
class CapsulePayload:
    """Stores information about a specific capsule payload.

    Attributes:
        payload (UefiCapsuleHeaderClass): an instance of UefiCapsuleHeaderClass that represents the payload data.
        payload_filename (str):  the payload filename as a string
        esrt_guid (uuid.UUID): the payload ESRT guid as a uuid.UUID instance.
        version (int): the 32-bit ESRT version for the payload.
        firmware_description (str): the firmware payload description.
        tag (str):  a uniquely identifying the payload. optional, if not present, will be auto-generated.
        rollback (bool): indicates whether this is a rollback payload. optional, defaults to false.
        integrity_data (bytes):  integrity data for this payload. optional.
        integrity_filename (str): integrity filename. optional if integrity_data is None, required otherwise.
    """
    payload: UefiCapsuleHeaderClass
    payload_filename: str
    esrt_guid: uuid.UUID
    version: int
    firmware_description: str
    tag: str = None
    rollback: bool = False
    integrity_data: bytes = None
    integrity_filename: str = None


@dataclass
class Capsule:
    """Stores information about a capsule (potentially with multiple payloads).

    Attributes:
        version_string (str): the version of the entire capsule driver package as a string (e.g. 1.0.0.1)
        name (str): the name of the capsule package
        provider_name (str): the name of the capsule provider
        arch (str): the architecture targeted by the capsule
        os (str): the OS targeted by the capsule.
        manufacturer_name (str): name of the capsule manufacturer. optional, defaults to provider_name if None.
        date (datetime.date): when the capsule was built. optional, defaults to datetime.date.today().
        payloads (List[CapsulePayload]): a list of capsule payloads. optional, defaults to empty list
    """
    version_string: str
    name: str
    provider_name: str
    arch: str = None
    os: str = None
    manufacturer_name: str = None
    date: datetime.date = datetime.date.today()
    payloads: List[CapsulePayload] = field(default_factory=list)


def get_capsule_file_name(capsule_options: dict) -> str:
    """Returns the formatted capsule file name from the capsule_options dict."""
    return f"{capsule_options['fw_name']}_{capsule_options['fw_version_string']}.bin"


def get_normalized_version_string(version_string: str) -> str:
    """Normalizes a version string that is compatible with INF and CAT files."""
    # 19H1 HLK requires a 4 digit version string, or it will fail
    while (version_string.count('.') < 3):
        version_string += '.0'
    return version_string


def get_default_arch() -> str:
    """Consistently return the default architecture for windows files."""
    return 'amd64'


def get_default_os_string() -> str:
    """Consistently return the default os for windows files."""
    return 'Win10'


def build_capsule(capsule_data: bytes, capsule_options: dict, signer_module: object,
                  signer_options: dict) -> UefiCapsuleHeaderClass:
    """Goes through all steps of capsule generation for a single-payload FMP capsule.

    takes in capsule_data as a byte string, a signer module, and capsule and signer options,
    and produces all of the headers necessary. Will use the signer module to produce the cert data
    for the FMP Auth header.

    NOTE: Uses a fixed MonotonicCount of 1.

    Args:
        capsule_data (bytes): innermost payload
        capsule_options (dict): contains the capsule payload fields. Must include
            'fw_version', 'lsv_version', and 'esrt_guid' at minimum. These should
            all be strings and the two versions should be strings of hex number
        signer_module (object): a capsule signer module that implements the sign() function
            (see pyopenssl_signer or signtool_signer build-in modules for examples)
        signer_options (dict): options that will be passed to the signer_module.

    Returns:
        (UefiCapsuleHeaderClass): the capsule header containing all data provided.
    """
    # Start building the capsule as we go.
    # Create the FMP Payload and set all the necessary options.
    fmp_payload_header = FmpPayloadHeaderClass()
    fmp_payload_header.FwVersion = int(capsule_options['fw_version'], 16)
    fmp_payload_header.LowestSupportedVersion = int(capsule_options['lsv_version'], 16)
    fmp_payload_header.Payload = capsule_data

    # Create the auth header and get ready to sign the data.
    fmp_auth_header = FmpAuthHeaderClass()
    fmp_auth_header.MonotonicCount = 1
    fmp_auth_header.FmpPayloadHeader = fmp_payload_header

    data_to_sign = fmp_payload_header.Encode()
    data_to_sign = data_to_sign + struct.pack("<Q", fmp_auth_header.MonotonicCount)

    # Sign the data and assign it to the cert data.
    signature_options = {
        'sign_alg': 'pkcs12',
        'hash_alg': 'sha256'
    }
    # Set or override OID.
    signer_options['oid'] = PKCS7_SIGNED_DATA_OID
    fmp_auth_header.AuthInfo.CertData = signer_module.sign(data_to_sign, signature_options, signer_options)

    fmp_capsule_image_header = FmpCapsuleImageHeaderClass()
    fmp_capsule_image_header.UpdateImageTypeId = uuid.UUID(capsule_options['esrt_guid'])
    fmp_capsule_image_header.UpdateImageIndex = 1
    fmp_capsule_image_header.FmpAuthHeader = fmp_auth_header

    fmp_capsule_header = FmpCapsuleHeaderClass()
    fmp_capsule_header.AddFmpCapsuleImageHeader(fmp_capsule_image_header)

    uefi_capsule_header = UefiCapsuleHeaderClass()
    uefi_capsule_header.FmpCapsuleHeader = fmp_capsule_header
    uefi_capsule_header.PersistAcrossReset = True
    uefi_capsule_header.InitiateReset = True

    return uefi_capsule_header


def save_capsule(uefi_capsule_header: UefiCapsuleHeaderClass, capsule_options: dict, save_path: str) -> str:
    """Serializes the capsule object to the target directory.

    takes in a UefiCapsuleHeaderClass object, a dictionary of capsule_options, and a filesystem directory path
    and serializes the capsule object to the target directory

    Args:
        uefi_capsule_header (UefiCapsuleHeaderClass): header to encode
        capsule_options (dict): used to get the capsule file name
        save_path (str): where to save the file to.

    Returns:
        (str): path the file was saved to.
    """
    # Expand the version string prior to creating the payload file.
    capsule_options['fw_version_string'] = get_normalized_version_string(capsule_options['fw_version_string'])

    # First, create the entire save path.
    os.makedirs(save_path, exist_ok=True)

    # Then save the file.
    capsule_file_path = os.path.join(save_path, get_capsule_file_name(capsule_options))
    with open(capsule_file_path, 'wb') as capsule_file:
        capsule_file.write(uefi_capsule_header.Encode())

    return capsule_file_path


def save_multinode_capsule(capsule: Capsule, save_path: str) -> str:
    """Generates capsule files from capsule object.

    takes in a Capsule object and a filesystem directory path and generates the capsule files at that path.

    Args:
        capsule (Capsule): a Capsule object containing the capsule details.
        save_path (str): directory path to save the capsule contents into

    Returns:
        (str): the save path
    """
    os.makedirs(save_path, exist_ok=True)
    for capsule_payload in capsule.payloads:
        payload_file_path = os.path.join(save_path, capsule_payload.payload_filename)
        with open(payload_file_path, 'wb') as payload_file:
            payload_file.write(capsule_payload.payload.Encode())

        if capsule_payload.integrity_data is not None:
            if (capsule_payload.integrity_filename is None):
                raise ValueError("Integrity data specified, but no integrity filename specified.")
            integrity_file_path = os.path.join(save_path, capsule_payload.integrity_filename)
            with open(integrity_file_path, 'wb') as integrity_file:
                integrity_file.write(capsule_payload.integrity_data)
    return save_path


def create_inf_file(capsule_options: dict, save_path: str) -> str:
    """Creates the Windows INF file for the UEFI capsule.

    takes in a dictionary of capsule_options and creates the Windows INF file for the UEFI capsule according
    to the provided options

    Args:
        capsule_options (dict): options to create the INF file
        save_path (str): directory to save the file.

    Returns:
        (str): INF file path

    NOTE: will save the final file to the save_path with a name determined from the capsule_options
    """
    # Expand the version string prior to creating INF file.
    capsule_options['fw_version_string'] = get_normalized_version_string(capsule_options['fw_version_string'])

    # Deal with optional parameters when creating the INF file.
    capsule_options['is_rollback'] = capsule_options.get('is_rollback', False)
    capsule_options['arch'] = capsule_options.get('arch', get_default_arch())
    capsule_options['mfg_name'] = capsule_options.get('mfg_name', capsule_options['provider_name'])

    inf_file = inf_generator2.InfFile(
        capsule_options['fw_name'],
        capsule_options['fw_version_string'],
        datetime.date.today().strftime("%m/%d/%Y"),
        capsule_options['provider_name'],
        capsule_options['mfg_name'],
        capsule_options['arch']
    )

    inf_file.AddFirmware(
        "Firmware",
        capsule_options['fw_description'],
        capsule_options['esrt_guid'],
        capsule_options['fw_version'],
        get_capsule_file_name(capsule_options),
        Rollback=capsule_options['is_rollback'],
        IntegrityFile=capsule_options.get('fw_integrity_file', None)
    )

    inf_file_path = os.path.join(save_path, f"{capsule_options['fw_name']}.inf")
    with open(inf_file_path, "w") as fp:
        fp.write(str(inf_file))

    return inf_file_path


def create_multinode_inf_file(capsule: Capsule, save_path: str) -> str:
    """Creates the Windows INF file from a capsule object.

    Takes in a capsule object containing payload information and creates the Windows INF file in save_path

    Args:
        capsule (Capsule): capsule object containing payload information
        save_path (str): path to directory where inf file will be created

    Returns:
        name of created inf file
    """
    # Expand the version string prior to creating INF file.
    capsule.version_string = get_normalized_version_string(capsule.version_string)

    # set defaults for non-specified fields
    if (capsule.arch is None):
        capsule.arch = get_default_arch()
    if (capsule.manufacturer_name is None):
        capsule.manufacturer_name = capsule.provider_name

    inf_file = inf_generator2.InfFile(
        capsule.name,
        capsule.version_string,
        capsule.date.strftime("%m/%d/%Y"),
        capsule.provider_name,
        capsule.manufacturer_name,
        capsule.arch
    )

    idx = 0
    for payload in capsule.payloads:
        if payload.tag is None:
            payload.tag = f"Firmware{idx}"
            idx += 1
        inf_file.AddFirmware(
            payload.tag,
            payload.firmware_description,
            str(payload.esrt_guid),
            str(payload.version),
            payload.payload_filename,
            Rollback=payload.rollback,
            IntegrityFile=payload.integrity_filename
        )

    inf_file_path = os.path.join(save_path, f"{capsule.name}.inf")
    with open(inf_file_path, "w") as fp:
        fp.write(str(inf_file))

    return inf_file_path


def create_cat_file(capsule_options: dict, save_path: str) -> str:
    """Creates the Windows CAT file for the UEFI capsule.

    Takes in a dictionary of capsule_options and creates the Windows CAT file for the UEFI capsule according
    to the provided options

    Args:
        capsule_options (dict): options to create the cat file
        save_path (str): directory to save file.

    Returns:
        (str): cat file path

    NOTE: will save the final file to the save_path with a name determined from the capsule_options
    """
    # Deal with optional parameters when creating the CAT file.
    capsule_options['arch'] = capsule_options.get('arch', get_default_arch())
    capsule_options['os_string'] = capsule_options.get('os_string', get_default_os_string())

    # Create the CAT.
    catgenerator = cat_generator.CatGenerator(
        capsule_options['arch'],
        capsule_options['os_string']
    )
    cat_file_path = os.path.join(save_path, f"{capsule_options['fw_name']}.cat")
    ret = catgenerator.MakeCat(cat_file_path)
    if (ret != 0):
        raise RuntimeError("MakeCat Failed with errorcode %d!" % ret)

    return cat_file_path
